from abc import ABC, abstractmethod
from typing import Optional, Union

import pandas as pd

from player_performance_ratings.ratings.rating_calculators import MatchRatingGenerator
from player_performance_ratings.ratings.enums import (
    RatingKnownFeatures,
    RatingHistoricalFeatures,
)

from player_performance_ratings.data_structures import (
    Match,
    PlayerRating,
    TeamRating,
    ColumnNames,
)


class RatingGenerator(ABC):

    def __init__(
        self,
        performance_column: str,
        non_estimator_known_features_out: Optional[list[RatingKnownFeatures]],
        historical_features_out: Optional[list[RatingHistoricalFeatures]],
        match_rating_generator: MatchRatingGenerator,
        seperate_player_by_position: Optional[bool] = False,
        prefix: str = "",
    ):
        self.performance_column = performance_column
        self.seperate_player_by_position = seperate_player_by_position
        self.match_rating_generator = match_rating_generator
        self._known_features_out = []
        self._historical_features_out = historical_features_out or []
        self.non_estimator_known_features_out = non_estimator_known_features_out
        self.prefix = prefix
        self.column_names = None
        self._calculated_match_ids = []

    def reset_ratings(self):
        self._calculated_match_ids = []

    @abstractmethod
    def generate_historical_by_matches(
        self,
        matches: list[Match],
        column_names: ColumnNames,
        historical_features_out: Optional[list[RatingHistoricalFeatures]] = None,
        known_features_out: Optional[list[RatingKnownFeatures]] = None,
    ) -> dict[Union[RatingKnownFeatures, RatingHistoricalFeatures], list[float]]:
        pass

    @abstractmethod
    def generate_historical(
        self,
        df: pd.DataFrame,
        column_names: ColumnNames,
        historical_features_out: Optional[list[RatingHistoricalFeatures]] = None,
        known_features_out: Optional[list[RatingKnownFeatures]] = None,
    ) -> pd.DataFrame:
        pass

    @abstractmethod
    def generate_future(
        self,
        df: Optional[pd.DataFrame],
        matches: Optional[list[Match]] = None,
        historical_features_out: Optional[list[RatingHistoricalFeatures]] = None,
        known_features_out: Optional[list[RatingKnownFeatures]] = None,
    ) -> pd.DataFrame:
        pass

    @property
    def known_features_out(self) -> list[str]:
        return self._known_features_out

    @property
    def features_out(
        self,
    ) -> list[str]:
        return [*self.known_features_out, *self.historical_features_out]

    @property
    def historical_features_out(self) -> list[str]:
        if self._historical_features_out:
            return self._historical_features_out
        return []

    @property
    def known_features_return(self) -> list[str]:
        if self.non_estimator_known_features_out:
            return list(
                set(self.non_estimator_known_features_out + self.known_features_out)
            )
        return self.known_features_out

    @property
    def player_ratings(self) -> dict[str, PlayerRating]:
        return dict(
            sorted(
                self.match_rating_generator.player_ratings.items(),
                key=lambda item: item[1].rating_value,
                reverse=True,
            )
        )

    @property
    def team_ratings(self) -> list[TeamRating]:
        team_id_ratings: list[TeamRating] = []
        teams = self.match_rating_generator.teams
        player_ratings = self.player_ratings
        for id, team in teams.items():
            team_player_ratings = [player_ratings[p] for p in team.player_ids]
            team_rating_value = sum(
                [p.rating_value for p in team_player_ratings]
            ) / len(team_player_ratings)
            team_id_ratings.append(
                TeamRating(
                    id=team.id,
                    name=team.name,
                    players=team_player_ratings,
                    last_match_day_number=team.last_match_day_number,
                    rating_value=team_rating_value,
                )
            )

        return list(
            sorted(team_id_ratings, key=lambda team: team.rating_value, reverse=True)
        )

    @property
    def calculated_match_ids(self) -> list[str]:
        return self._calculated_match_ids

    def _validate_match(self, match: Match):
        if len(match.teams) < 2:
            raise ValueError(f"{match.id} only contains {len(match.teams)} teams")



import copy
from dataclasses import dataclass

from typing import Optional, Union

import optuna
import pandas as pd
from optuna.samplers import TPESampler
from optuna.trial import BaseTrial

from player_performance_ratings.cross_validator.cross_validator import CrossValidator
from player_performance_ratings.ratings.performance_generator import (
    ColumnWeight,
    PerformancesGenerator,
)
from player_performance_ratings.ratings.enums import RatingKnownFeatures

from player_performance_ratings import PipelineFactory

from player_performance_ratings.tuner.utils import (
    add_params_from_search_range,
    ParameterSearchRange,
)

RC = RatingKnownFeatures


@dataclass
class PerformancesSearchRange:
    search_ranges: list[ParameterSearchRange]
    name: str = "performance"


class PerformancesGeneratorTuner:

    def __init__(
        self,
        performances_search_range: Union[
            list[PerformancesSearchRange], PerformancesSearchRange
        ],
        feature_names: Optional[Union[list[str], list[list[str]]]] = None,
        n_trials: int = 30,
    ):

        self.performances_search_ranges = (
            performances_search_range
            if isinstance(performances_search_range, list)
            else [performances_search_range]
        )
        self.feature_names = feature_names

        self.n_trials = n_trials
        self._scores = []

    def tune(
        self,
        df: pd.DataFrame,
        rating_idx: int,
        cross_validator: CrossValidator,
        pipeline_factory: PipelineFactory,
    ) -> PerformancesGenerator:

        if pipeline_factory.performances_generator is None:
            raise ValueError(
                "pipeline_factory.performances_generator is None. Please provide a performances_generator."
            )

        df = df.copy()

        pipeline_factory = copy.deepcopy(pipeline_factory)

        def objective(
            trial: BaseTrial,
            df: pd.DataFrame,
            pipeline_factory: PipelineFactory,
        ) -> float:

            best_pre_transformers = [
                copy.deepcopy(p)
                for p in pipeline_factory.performances_generator.original_transformers
            ]
            performances = [
                cw for cw in pipeline_factory.performances_generator.performances
            ]
            for performance_search_range in self.performances_search_ranges:
                performance_name = performance_search_range.name
                search_range = performance_search_range.search_ranges
                performances[rating_idx].weights = []
                raw_params = {}
                renamed_search = copy.deepcopy(search_range)
                for sr in renamed_search:
                    sr.name = f"{performance_name}__{sr.name}"
                add_params_from_search_range(
                    trial=trial,
                    params=raw_params,
                    parameter_search_range=renamed_search,
                )
                new_col_weights = self._create_column_weights(
                    params=raw_params,
                    remove_string=f"{performance_name}__",
                    search_range=search_range,
                )
                for new_col_weight in new_col_weights:
                    if new_col_weight.name in [
                        cw.name for cw in performances[rating_idx].weights
                    ]:
                        performances[rating_idx].weights[
                            [cw.name for cw in performances[rating_idx].weights].index(
                                new_col_weight.name
                            )
                        ] = new_col_weight
                    else:
                        performances[rating_idx].weights.append(new_col_weight)

            performances_generator = PerformancesGenerator(
                performances=performances,
                transformers=best_pre_transformers,
                auto_transform_performance=pipeline_factory.performances_generator.auto_transform_performance,
            )
            pipeline = pipeline_factory.create(
                performances_generator=performances_generator
            )
            return pipeline.cross_validate_score(
                df=df,
                cross_validator=cross_validator,
                create_performance=True,
                create_rating_features=True,
            )

        direction = "minimize"
        study_name = "optuna_study"
        optuna_seed = 12
        sampler = TPESampler(seed=optuna_seed)
        study = optuna.create_study(
            direction=direction, study_name=study_name, sampler=sampler
        )
        callbacks = []
        study.optimize(
            lambda trial: objective(trial, df=df, pipeline_factory=pipeline_factory),
            n_trials=self.n_trials,
            callbacks=callbacks,
        )

        best_params = study.best_params
        best_column_weights = self._select_best_performances(all_params=best_params)
        performances = [
            cw for cw in pipeline_factory.performances_generator.performances
        ]
        performances[rating_idx].weights = best_column_weights[rating_idx]
        return PerformancesGenerator(
            performances=performances,
            transformers=pipeline_factory.performances_generator.original_transformers,
        )

    def _create_column_weights(
        self, params: dict, remove_string: str, search_range: list[ParameterSearchRange]
    ) -> list[ColumnWeight]:

        sum_weights = sum([v for _, v in params.items()])
        column_weights = []
        lower_is_better = False
        for name, weight in params.items():
            for search in search_range:
                if search.name == name.split("__")[1]:
                    lower_is_better = search.lower_is_better
                    break
            column_weights.append(
                ColumnWeight(
                    name=name.replace(remove_string, ""),
                    weight=weight / sum_weights,
                    lower_is_better=lower_is_better,
                )
            )

        return column_weights

    def _select_best_performances(self, all_params: dict) -> list[list[ColumnWeight]]:
        best_column_weights = []
        for performance_search_range in self.performances_search_ranges:
            performance_name = performance_search_range.name
            search_range = performance_search_range.search_ranges
            column_weights = []
            sum_weights = 0
            for param in all_params:

                if f"{performance_name}__" in param:
                    sum_weights += all_params[param]

            for param in all_params:

                if f"{performance_name}__" in param:
                    lower_is_better = False
                    for s in search_range:
                        if s.name == param.split("__")[1]:
                            lower_is_better = s.lower_is_better
                            break

                    column_weights.append(
                        ColumnWeight(
                            name=param.replace(f"{performance_name}__", ""),
                            weight=all_params[param] / sum_weights,
                            lower_is_better=lower_is_better,
                        )
                    )

            best_column_weights.append(column_weights)

        return best_column_weights



import math
from dataclasses import dataclass
from abc import ABC, abstractmethod
from enum import Enum
from typing import Optional, Callable, Union, Any

import numpy as np
import pandas as pd

from player_performance_ratings.consts import PredictColumnNames


class Operator(Enum):
    EQUALS = "=="
    NOT_EQUALS = "!="
    GREATER_THAN = ">"
    LESS_THAN = "<"
    GREATER_THAN_OR_EQUALS = ">="
    LESS_THAN_OR_EQUALS = "<="
    IN = "in"
    NOT_IN = "not in"


@dataclass
class Filter:
    column_name: str
    value: Union[Any, list[Any]]
    operator: Operator


def apply_filters(df: pd.DataFrame, filters: list[Filter]) -> pd.DataFrame:
    df = df.copy()
    for filter in filters:
        if filter.operator == Operator.EQUALS:
            df = df[df[filter.column_name] == filter.value]
        elif filter.operator == Operator.NOT_EQUALS:
            df = df[df[filter.column_name] != filter.value]
        elif filter.operator == Operator.GREATER_THAN:
            df = df[df[filter.column_name] > filter.value]
        elif filter.operator == Operator.LESS_THAN:
            df = df[df[filter.column_name] < filter.value]
        elif filter.operator == Operator.GREATER_THAN_OR_EQUALS:
            df = df[df[filter.column_name] >= filter.value]
        elif filter.operator == Operator.LESS_THAN_OR_EQUALS:
            df = df[df[filter.column_name] <= filter.value]
        elif filter.operator == Operator.IN:
            df = df[df[filter.column_name].isin(filter.value)]
        elif filter.operator == Operator.NOT_IN:
            df = df[~df[filter.column_name].isin(filter.value)]

    return df


class BaseScorer(ABC):

    def __init__(
            self,
            target: str,
            pred_column: str,
            validation_column: Optional[str],
            filters: Optional[list[Filter]] = None,
            granularity: Optional[list[str]] = None,
    ):
        self.target = target
        self.pred_column = pred_column
        self.validation_column = validation_column
        self.filters = filters or []
        self.granularity = granularity

    @abstractmethod
    def score(self, df: pd.DataFrame) -> float:
        pass


class SklearnScorer(BaseScorer):

    def __init__(
            self,
            pred_column: str,
            scorer_function: Callable,
            target: Optional[str] = PredictColumnNames.TARGET,
            validation_column: Optional[str] = None,
            granularity: Optional[list[str]] = None,
            filters: Optional[list[Filter]] = None,
    ):
        """
        :param pred_column: The column name of the predictions
        :param scorer_function: SKlearn scorer function, e.g. los_loss
        :param target: The column name of the target
        :param validation_column: The column name of the validation column.
            If set, the scorer will be calculated only once the values of the validation column are equal to 1
        :param granularity: The columns to group by before calculating the score
        :param filters: The filters to apply before calculating
        """

        self.pred_column_name = pred_column
        self.scorer_function = scorer_function
        super().__init__(
            target=target,
            pred_column=pred_column,
            granularity=granularity,
            filters=filters,
            validation_column=validation_column,
        )

    def score(self, df: pd.DataFrame) -> float:
        df = df.copy()
        if self.validation_column:
            filters = self.filters or [] + [Filter(self.validation_column, 1, Operator.EQUALS)]
        else:
            filters = self.filters or []
        df = apply_filters(df, filters)
        if self.granularity:
            grouped = (
                df.groupby(self.granularity)[self.pred_column_name, self.target]
                .mean()
                .reset_index()
            )
        else:
            grouped = df
        if isinstance(df[self.pred_column_name].iloc[0], list):
            return self.scorer_function(
                grouped[self.target],
                np.asarray(grouped[self.pred_column_name]).tolist(),
            )
        return self.scorer_function(
            grouped[self.target].astype(int), grouped[self.pred_column_name]
        )


class OrdinalLossScorer(BaseScorer):

    def __init__(
            self,
            pred_column: str,
            class_column_name: str = "classes",
            target: Optional[str] = PredictColumnNames.TARGET,
            validation_column: Optional[str] = None,
            granularity: Optional[list[str]] = None,
            filters: Optional[list[Filter]] = None,
    ):
        """
        :param pred_column: The column name of the predictions
        :param class_column_name: The column name that contains the integer values of the classes.
        :param scorer_function: SKlearn scorer function, e.g. los_loss
        :param target: The column name of the target
        :param validation_column: The column name of the validation column.
            If set, the scorer will be calculated only once the values of the validation column are equal to 1
        :param granularity: The columns to group by before calculating the score
        :param filters: The filters to apply before calculating
        """

        self.pred_column_name = pred_column
        self.class_column_name = class_column_name
        self.granularity = granularity
        super().__init__(
            target=target,
            pred_column=pred_column,
            filters=filters,
            granularity=granularity,
            validation_column=validation_column,
        )

    def score(self, df: pd.DataFrame) -> float:

        df = df.copy()
        df = apply_filters(df, self.filters)
        df.reset_index(drop=True, inplace=True)

        distinct_classes_variations = df.drop_duplicates(subset=[self.class_column_name])[
            self.class_column_name
        ].tolist()

        sum_lrs = [0 for _ in range(len(distinct_classes_variations))]
        sum_lr = 0
        for variation_idx, distinct_class_variation in enumerate(
                distinct_classes_variations
        ):

            if not isinstance(distinct_class_variation, list):
                if math.isnan(distinct_class_variation):
                    continue

            rows_target_group = df[
                df[self.class_column_name].apply(lambda x: x == distinct_class_variation)
            ]
            probs = rows_target_group[self.pred_column_name]
            last_column_name = f"prob_under_{distinct_class_variation[0] - 0.5}"
            rows_target_group[last_column_name] = probs.apply(lambda x: x[0])

            for idx, class_ in enumerate(distinct_class_variation[1:]):

                p_c = "prob_under_" + str(class_ + 0.5)
                rows_target_group[p_c] = (
                        probs.apply(lambda x: x[idx + 1])
                        + rows_target_group[last_column_name]
                )

                count_exact = len(
                    rows_target_group[rows_target_group["__target"] == class_]
                )
                weight_class = count_exact / len(rows_target_group)

                if self.granularity:
                    grouped = (
                        rows_target_group.groupby(self.granularity + ["__target"])[p_c]
                        .mean()
                        .reset_index()
                    )
                else:
                    grouped = rows_target_group

                grouped["min"] = 0.0001
                grouped["max"] = 0.9999
                grouped[p_c] = np.minimum(grouped["max"], grouped[p_c])
                grouped[p_c] = np.maximum(grouped["min"], grouped[p_c])
                grouped["log_loss"] = 0
                grouped.loc[grouped["__target"] <= class_, "log_loss"] = np.log(
                    grouped[p_c]
                )
                grouped.loc[grouped["__target"] > class_, "log_loss"] = np.log(
                    1 - grouped[p_c]
                )
                log_loss = grouped["log_loss"].mean()
                sum_lrs[variation_idx] -= log_loss * weight_class

                last_column_name = p_c
            sum_lr += sum_lrs[variation_idx] * len(rows_target_group) / len(df)
        return sum_lr



import pandas as pd
from sklearn.metrics import log_loss

from player_performance_ratings import PredictColumnNames

from player_performance_ratings.pipeline import Pipeline
from player_performance_ratings.predictor import GameTeamPredictor

from player_performance_ratings.ratings import UpdateRatingGenerator

from player_performance_ratings.data_structures import ColumnNames
from player_performance_ratings.scorer import SklearnScorer

df = pd.read_pickle("data/game_player_subsample.pickle")

# Defines the column names as they appear in the dataframe
column_names = ColumnNames(
    team_id="team_id",
    match_id="game_id",
    start_date="start_date",
    player_id="player_name",
)
# Sorts the dataframe. The dataframe must always be sorted as below
df = df.sort_values(
    by=[
        column_names.start_date,
        column_names.match_id,
        column_names.team_id,
        column_names.player_id,
    ]
)

# Defines the target column we inted to predict
df[PredictColumnNames.TARGET] = df["won"].astype(int)

# Drops games with less or more than 2 teams
df = (
    df.assign(
        team_count=df.groupby(column_names.match_id)[column_names.team_id].transform(
            "nunique"
        )
    )
    .loc[lambda x: x.team_count == 2]
    .drop(columns=["team_count"])
)

# Pretends the last 10 games are future games. The most will be trained on everything before that.
most_recent_10_games = df[column_names.match_id].unique()[-10:]
historical_df = df[~df[column_names.match_id].isin(most_recent_10_games)]
future_df = df[df[column_names.match_id].isin(most_recent_10_games)].drop(
    columns=[PredictColumnNames.TARGET, "won"]
)

# Defining a simple rating-generator. It will use the "won" column to update the ratings.
# In contrast to a typical Elo, ratings will follow players.
rating_generator = UpdateRatingGenerator(performance_column="won")

# Defines the predictor. A machine-learning model will be used to predict game winner on a game-team-level.
# Mean team-ratings will be calculated (from player-level) and rating-difference between the 2 teams calculated.
# It will also use the location of the game as a feature.
predictor = GameTeamPredictor(
    game_id_colum=column_names.match_id,
    team_id_column=column_names.team_id,
    estimator_features=["location"],
)

# Pipeline is whether we define all the steps. Other transformations can take place as well.
# However, in our simple example we only have a simple rating-generator and a predictor.
pipeline = Pipeline(
    rating_generators=rating_generator,
    predictor=predictor,
    column_names=column_names,
)

# Trains the model and returns historical predictions
historical_predictions = pipeline.train_predict(
    df=historical_df, cross_validate_predict=True
)


historical_grouped_predictions = historical_predictions.groupby(column_names.match_id).first()[
    [
        column_names.start_date,
        column_names.team_id,
        "team_id_opponent",
        predictor.pred_column,
        predictor.target,
        "is_validation"
    ]
].reset_index()
scorer = SklearnScorer(pred_column=predictor.pred_column, target=predictor.target, scorer_function=log_loss,
                       validation_column="is_validation")

score = scorer.score(df=historical_grouped_predictions)
print(f"Logloss Score on historical data: {score}")


# Future predictions on future results
future_predictions = pipeline.future_predict(df=future_df)

# Grouping predictions from game-player level to game-level.
team_grouped_predictions = future_predictions.groupby(column_names.match_id).first()[
    [
        column_names.start_date,
        column_names.team_id,
        "team_id_opponent",
        predictor.pred_column,

    ]
]

print(team_grouped_predictions)



from abc import abstractmethod, ABC
from typing import Optional

import pandas as pd
from player_performance_ratings import PredictColumnNames
import warnings

warnings.simplefilter(action="ignore", category=pd.errors.PerformanceWarning)


class PredictorTransformer(ABC):

    def __init__(self, features: list[str]):
        self.features = features

    @abstractmethod
    def fit_transform(self, df: pd.DataFrame) -> pd.DataFrame:
        pass

    @abstractmethod
    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        pass

    @property
    @abstractmethod
    def features_out(self) -> list[str]:
        pass


class ConvertDataFrameToCategoricalTransformer(PredictorTransformer):
    """
    Converts a specified list of columns to categorical dtype
    """

    def __init__(self, features: list[str]):
        super().__init__(features=features)

    def fit_transform(self, df: pd.DataFrame) -> pd.DataFrame:
        self._features_out = self.features
        return self.transform(df)[self._features_out]

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        for feature in self.features:
            df = df.assign(**{feature: df[feature].astype("category")})
        return df[self._features_out]

    @property
    def features_out(self) -> list[str]:
        return self._features_out


class SkLearnTransformerWrapper(PredictorTransformer):
    """
    A wrapper around an Sklearn Transformer
    """

    def __init__(self, transformer, features: list[str]):
        self.transformer = transformer
        super().__init__(features=features)
        self._features_out = []

    def fit_transform(self, df: pd.DataFrame) -> pd.DataFrame:

        try:
            transformed_values = self.transformer.fit_transform(
                df[self.features]
            ).toarray()
        except AttributeError:
            transformed_values = self.transformer.fit_transform(df[self.features])
            if isinstance(transformed_values, pd.DataFrame):
                transformed_values = transformed_values.to_numpy()

        self._features_out = self.transformer.get_feature_names_out().tolist()

        return df.assign(
            **{
                self._features_out[idx]: transformed_values[:, idx]
                for idx in range(len(self._features_out))
            }
        )[self._features_out]

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        try:
            transformed_values = self.transformer.transform(df[self.features]).toarray()
        except AttributeError:
            transformed_values = self.transformer.transform(df[self.features])
            if isinstance(transformed_values, pd.DataFrame):
                transformed_values = transformed_values.to_numpy()
        return df.assign(
            **{
                self._features_out[idx]: transformed_values[:, idx]
                for idx in range(len(self._features_out))
            }
        )[self._features_out]

    @property
    def features_out(self) -> list[str]:
        return self._features_out



from typing import Optional, Union

import numpy as np
import pandas as pd
import polars as pl

from player_performance_ratings import ColumnNames
from player_performance_ratings.transformers.base_transformer import (
    BaseLagGenerator,
    BaseLagGeneratorPolars,
)
from player_performance_ratings.utils import validate_sorting


def create_output_column_by_game_group(
    data: pd.DataFrame,
    feature_name: str,
    weight_column: str,
    game_id: str,
    granularity: list[str],
) -> pd.DataFrame:
    if weight_column:
        data = data.assign(
            **{f"weighted_{feature_name}": data[feature_name] * data[weight_column]}
        )
        data = data.assign(
            **{
                "sum_weighted": data.groupby(granularity + [game_id])[
                    weight_column
                ].transform("sum")
            }
        )
        data = data.assign(
            **{feature_name: data[f"weighted_{feature_name}"] / data["sum_weighted"]}
        )
        data = data.groupby(granularity + [game_id])[feature_name].sum().reset_index()

    else:
        data = data.groupby(granularity + [game_id])[feature_name].mean().reset_index()
    return data


class LagTransformer(BaseLagGenerator):

    def __init__(
        self,
        features: list[str],
        lag_length: int,
        granularity: Optional[list[str]] = None,
        days_between_lags: Optional[list[int]] = None,
        future_lag: bool = False,
        prefix: str = "lag_",
        add_opponent: bool = False,
    ):
        """
        :param features. List of features to create lags for
        :param lag_length: Number of lags
        :param granularity: Columns to group by before lagging. E.g. player_id or [player_id, position].
            In the latter case it will get the lag for each player_id and position combination.
            Defaults to
        :param days_between_lags. Adds a column with the number of days between the lagged date and the current date.
        :param future_lag: If True, the lag will be calculated for the future instead of the past.
        :param prefix:
            Prefix for the new lag columns
        """

        super().__init__(
            features=features,
            add_opponent=add_opponent,
            prefix=prefix,
            iterations=[i for i in range(1, lag_length + 1)],
            granularity=granularity,
        )
        self.days_between_lags = days_between_lags or []
        for days_lag in self.days_between_lags:
            self._features_out.append(f"{prefix}{days_lag}_days_ago")

        self.lag_length = lag_length
        self.future_lag = future_lag
        self._df = None

    def generate_historical(
        self, df: pd.DataFrame, column_names: ColumnNames
    ) -> pd.DataFrame:
        """ """
        df = df.assign(is_future=0)
        self.column_names = column_names
        self.granularity = self.granularity or [self.column_names.player_id]
        validate_sorting(df=df, column_names=self.column_names)
        self._store_df(df)
        concat_df = self._generate_concat_df_with_feats(df)
        df = self._create_transformed_df(df=df, concat_df=concat_df)
        if "is_future" in df.columns:
            df = df.drop(columns="is_future")
        return df

    def generate_future(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.assign(is_future=1)
        concat_df = self._generate_concat_df_with_feats(df=df)
        transformed_df = concat_df[
            concat_df[self.column_names.match_id].isin(
                df[self.column_names.match_id].astype("str").unique().tolist()
            )
        ]
        transformed_future = self._generate_future_feats(
            transformed_df=transformed_df, ori_df=df
        )
        if "is_future" in transformed_future.columns:
            transformed_future = transformed_future.drop(columns="is_future")
        return transformed_future

    def _generate_concat_df_with_feats(self, df: pd.DataFrame) -> pd.DataFrame:

        if self._df is None:
            raise ValueError("fit_transform needs to be called before transform")

        if len(
            df.drop_duplicates(
                subset=[
                    self.column_names.player_id,
                    self.column_names.team_id,
                    self.column_names.match_id,
                ]
            )
        ) != len(df):
            raise ValueError(
                f"Duplicated rows in df. Df must be a unique combination of {self.column_names.player_id} and {self.column_names.update_match_id}"
            )

        concat_df = self._concat_df(df=df)

        if self.column_names.participation_weight:
            for feature in self.features:
                concat_df = concat_df.assign(
                    **{
                        feature: concat_df[feature]
                        * concat_df[self.column_names.participation_weight]
                    }
                )

        grouped = (
            concat_df.groupby(
                self.granularity
                + [self.column_names.update_match_id, self.column_names.start_date]
            )[self.features]
            .mean()
            .reset_index()
        )

        grouped = grouped.sort_values(
            [self.column_names.start_date, self.column_names.update_match_id]
        )

        for days_lag in self.days_between_lags:
            if self.future_lag:
                grouped["shifted_days"] = grouped.groupby(self.granularity)[
                    self.column_names.start_date
                ].shift(-days_lag)
                grouped[f"{self.prefix}{days_lag}_days_ago"] = (
                    pd.to_datetime(grouped["shifted_days"])
                    - pd.to_datetime(grouped[self.column_names.start_date])
                ).dt.days
            else:
                grouped["shifted_days"] = grouped.groupby(self.granularity)[
                    self.column_names.start_date
                ].shift(days_lag)
                grouped[f"{self.prefix}{days_lag}_days_ago"] = (
                    pd.to_datetime(grouped[self.column_names.start_date])
                    - pd.to_datetime(grouped["shifted_days"])
                ).dt.days

            grouped = grouped.drop(columns=["shifted_days"])

        for feature_name in self.features:
            for lag in range(1, self.lag_length + 1):
                output_column_name = f"{self.prefix}{lag}_{feature_name}"
                if output_column_name in concat_df.columns:
                    raise ValueError(
                        f"Column {output_column_name} already exists. Choose different prefix or ensure no duplication was performed"
                    )

        for feature_name in self.features:
            for lag in range(1, self.lag_length + 1):
                output_column_name = f"{self.prefix}{lag}_{feature_name}"
                if self.future_lag:
                    grouped = grouped.assign(
                        **{
                            output_column_name: grouped.groupby(self.granularity)[
                                feature_name
                            ].shift(-lag)
                        }
                    )
                else:
                    grouped = grouped.assign(
                        **{
                            output_column_name: grouped.groupby(self.granularity)[
                                feature_name
                            ].shift(lag)
                        }
                    )

        feats_out = []
        for feature_name in self.features:
            for lag in range(1, self.lag_length + 1):
                feats_out.append(f"{self.prefix}{lag}_{feature_name}")

        for days_lag in self.days_between_lags:
            feats_out.append(f"{self.prefix}{days_lag}_days_ago")

        concat_df = concat_df.merge(
            grouped[
                self.granularity
                + [
                    self.column_names.update_match_id,
                    self.column_names.start_date,
                    *feats_out,
                ]
            ],
            on=self.granularity
            + [self.column_names.update_match_id, self.column_names.start_date],
            how="left",
        )

        return concat_df

    @property
    def features_out(self) -> list[str]:
        return self._features_out


class RollingMeanTransformer(BaseLagGenerator):
    """
    Calculates the rolling mean for a list of features over a window of matches.
    Rolling Mean Values are also shifted by one match to avoid data leakage.

    Use .generate_historical() to generate rolling mean for historical data.
        The historical data is stored as instance-variables after calling .generate_historical()
    Use .generate_future() to generate rolling mean for future data after having called .generate_historical()
    """

    def __init__(
        self,
        features: list[str],
        window: int,
        min_periods: int = 1,
        granularity: Union[list[str], str] = None,
        add_opponent: bool = False,
        are_estimator_features=True,
        prefix: str = "rolling_mean_",
    ):
        """
        Calculates the rolling mean for a list of features over a window of matches.
        Rolling Mean Values are also shifted by one match to avoid data leakage.

        :param features:   Features to create rolling mean for
        :param window: Window size for rolling mean.
            If 10 will calculate rolling mean over the prior 10 observations
        :param min_periods: Minimum number of observations in window required to have a non-null result
        :param granularity: Columns to group by before rolling mean. E.g. player_id or [player_id, position].
             In the latter case it will get the rolling mean for each player_id and position combination.
             Defaults to player_id
        :param add_opponent: If True, will add new columns containing rolling mean for the opponent team.
        :param are_estimator_features: If True, the features will be added to the estimator features.
            If false, it makes it possible for the outer layer (Pipeline) to exclude these features from the estimator.
        :param prefix: Prefix for the new rolling mean columns
        """
        super().__init__(
            features=features,
            add_opponent=add_opponent,
            iterations=[window],
            prefix=prefix,
            granularity=granularity,
            are_estimator_features=are_estimator_features,
        )
        self.window = window
        self.min_periods = min_periods

    def generate_historical(
        self, df: pd.DataFrame, column_names: ColumnNames
    ) -> pd.DataFrame:
        """
        Generates rolling mean for historical data
        Stored the historical data as instance-variables so it's possible to generate future data afterwards
        The calculation is done using Polars.
         However, Pandas Dataframe can be used as input and it will also output a pandas dataframe in that case.

        :param df: Historical data
        :param column_names: Column names
        """
        df = df.assign(is_future=0)
        self.column_names = column_names
        self.granularity = self.granularity or [self.column_names.player_id]
        validate_sorting(df=df, column_names=self.column_names)
        self._store_df(df)
        concat_df = self._generate_concat_df_with_feats(df)
        df = self._create_transformed_df(df=df, concat_df=concat_df)
        if "is_future" in df.columns:
            df = df.drop(columns="is_future")
        return df

    def generate_future(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Generates rolling mean for future data
        Assumes that .generate_historical() has been called before

        Regardless of the scheduled data of the future match, all future matches are perceived as being played in the same date.
        That is to ensure that a team's 2nd future match has the same rolling means as the 1st future match.
        :param df: Future data
        """

        df = df.assign(is_future=1)
        concat_df = self._generate_concat_df_with_feats(df=df)
        transformed_df = concat_df[
            concat_df[self.column_names.match_id].isin(
                df[self.column_names.match_id].astype("str").unique().tolist()
            )
        ]
        transformed_future = self._generate_future_feats(
            transformed_df=transformed_df, ori_df=df
        )
        if "is_future" in transformed_future.columns:
            transformed_future = transformed_future.drop(columns="is_future")
        return transformed_future

    def _generate_concat_df_with_feats(self, df: pd.DataFrame) -> pd.DataFrame:

        if self._df is None:
            raise ValueError("fit_transform needs to be called before transform")

        concat_df = self._concat_df(df)

        for feature_name in self.features:

            if self.column_names.participation_weight:
                concat_df = concat_df.assign(
                    **{
                        feature_name: concat_df[feature_name]
                        * concat_df[self.column_names.participation_weight]
                    }
                )

            output_column_name = f"{self.prefix}{self.window}_{feature_name}"
            if output_column_name in concat_df.columns:
                raise ValueError(
                    f"Column {output_column_name} already exists. Choose different prefix or ensure no duplication was performed"
                )

            agg_dict = {feature_name: "mean", self.column_names.start_date: "first"}
            grp = (
                concat_df.groupby(
                    self.granularity + [self.column_names.update_match_id]
                )
                .agg(agg_dict)
                .reset_index()
            )
            grp.sort_values(
                by=[self.column_names.start_date, self.column_names.update_match_id],
                inplace=True,
            )

            grp = grp.assign(
                **{
                    output_column_name: grp.groupby(self.granularity)[
                        feature_name
                    ].transform(
                        lambda x: x.shift()
                        .rolling(self.window, min_periods=self.min_periods)
                        .mean()
                    )
                }
            )

            concat_df = concat_df.merge(
                grp[
                    self.granularity
                    + [self.column_names.update_match_id, output_column_name]
                ],
                on=self.granularity + [self.column_names.update_match_id],
                how="left",
            )
            concat_df = concat_df.sort_values(
                by=[
                    self.column_names.start_date,
                    self.column_names.match_id,
                    self.column_names.team_id,
                    self.column_names.player_id,
                ]
            )

        feats_added = [f for f in self.features_out if f in concat_df.columns]

        concat_df[feats_added] = concat_df.groupby(self.granularity)[
            feats_added
        ].fillna(method="ffill")
        return concat_df

    @property
    def features_out(self) -> list[str]:
        return self._features_out


class RollingMeanDaysTransformer(BaseLagGenerator):

    def __init__(
        self,
        features: list[str],
        days: Union[int, list[int]],
        granularity: Union[list[str], str] = None,
        add_count: bool = False,
        add_opponent: bool = False,
        prefix: str = "rolling_mean_days_",
    ):
        self.days = days
        if isinstance(self.days, int):
            self.days = [self.days]
        super().__init__(
            features=features,
            iterations=[i for i in self.days],
            prefix=prefix,
            add_opponent=add_opponent,
            granularity=granularity,
        )

        self.add_count = add_count
        self._fitted_game_ids = []

        for day in self.days:
            if self.add_count:
                feature = f"{self.prefix}{day}_count"
                self._features_out.append(feature)
                self._entity_features.append(feature)

                if self.add_opponent:
                    self._features_out.append(f"{self.prefix}{day}_count_opponent")

    def generate_historical(
        self, df: pd.DataFrame, column_names: ColumnNames
    ) -> pd.DataFrame:
        ori_types = {c: df[c].dtype for c in df.columns}
        df = df.assign(is_future=0)
        self.column_names = column_names
        self.granularity = self.granularity or [self.column_names.player_id]
        validate_sorting(df=df, column_names=self.column_names)
        self._store_df(df)
        concat_df = self._generate_concat_df_with_feats(df)
        df = self._create_transformed_df(df=df, concat_df=concat_df)

        if self.add_count:
            for day in self.days:
                df = df.assign(
                    **{
                        f"{self.prefix}{day}_count": df[
                            f"{self.prefix}{day}_count"
                        ].fillna(0)
                    }
                )
                if self.add_opponent:
                    df = df.assign(
                        **{
                            f"{self.prefix}{day}_count_opponent": df[
                                f"{self.prefix}{day}_count_opponent"
                            ].fillna(0)
                        }
                    )
        if "is_future" in df.columns:
            df = df.drop(columns="is_future")

        for c, t in ori_types.items():
            if c not in self.features_out and df[c].dtype != t:
                df[c] = df[c].astype(t)
        return df

    def generate_future(self, df: pd.DataFrame) -> pd.DataFrame:

        ori_types = {c: df[c].dtype for c in df.columns}
        df = df.assign(is_future=1)
        concat_df = self._generate_concat_df_with_feats(df=df)
        transformed_df = concat_df[
            concat_df[self.column_names.match_id].isin(
                df[self.column_names.match_id].astype("str").unique().tolist()
            )
        ]
        transformed_future = self._generate_future_feats(
            transformed_df=transformed_df, ori_df=df
        )
        if "is_future" in transformed_future.columns:
            transformed_future = transformed_future.drop(columns="is_future")

        for c, t in ori_types.items():
            if c not in self.features_out and transformed_future[c].dtype != t:
                transformed_future[c] = transformed_future[c].astype(t)
        return transformed_future

    def _generate_concat_df_with_feats(self, df: pd.DataFrame) -> pd.DataFrame:
        if self._df is None:
            raise ValueError("fit_transform needs to be called before transform")

        concat_df = self._concat_df(df)

        for feature_name in self.features:

            if self.column_names.participation_weight:
                concat_df = concat_df.assign(
                    **{
                        feature_name: concat_df[feature_name]
                        * concat_df[self.column_names.participation_weight]
                    }
                )

        concat_df = concat_df.assign(
            **{"__date_day": lambda x: x[self.column_names.start_date].dt.date}
        )

        aggregations = {feature: "mean" for feature in self.features}
        aggregations["__date_day"] = "first"
        grouped = (
            concat_df.groupby([*self.granularity, self.column_names.update_match_id])
            .agg(aggregations)
            .reset_index()
        )
        count_feats = []
        for day in self.days:
            prefix_day = f"{self.prefix}{day}"
            if self.add_count:
                count_feats.append(f"{prefix_day}_count")
            grouped = self._add_rolling_feature(
                concat_df=grouped,
                day=day,
                granularity=self.granularity,
                prefix_day=prefix_day,
            )

        feats_created = [f for f in self.features_out if f in grouped.columns]
        concat_df = concat_df.merge(
            grouped[
                [*self.granularity, *feats_created, self.column_names.update_match_id]
            ],
            on=self.granularity + [self.column_names.update_match_id],
            how="left",
        )
        concat_df = concat_df.sort_values(
            by=[
                self.column_names.start_date,
                self.column_names.match_id,
                self.column_names.team_id,
                self.column_names.player_id,
            ]
        )
        return concat_df.drop(columns="__date_day")

    def _add_rolling_feature(
        self, concat_df: pd.DataFrame, day: int, granularity: list[str], prefix_day: str
    ):

        if len(granularity) > 1:
            granularity_concat = "__".join(granularity)
            temporary_str_df = concat_df[granularity].astype(str)
            concat_df[granularity_concat] = temporary_str_df.agg("__".join, axis=1)
        else:
            granularity_concat = granularity[0]

        concat_df["is_nan"] = concat_df[self.features[0]].isna().astype(float)

        df1 = (
            concat_df.groupby(["__date_day", granularity_concat])[
                [*self.features, "is_nan"]
            ]
            .agg(["sum", "size"])
            .unstack()
            .asfreq("d", fill_value=np.nan)
            .rolling(window=day, min_periods=1)
            .sum()
            .shift()
            .stack()
        )
        feats = []
        for feature_name in self.features:
            feats.append(f"{prefix_day}_{feature_name}_sum")
            feats.append(f"{prefix_day}_{feature_name}_count")

        df1.columns = feats + ["is_nan_sum", "is_nan_count"]
        for feature_name in self.features:
            df1[f"{prefix_day}_{feature_name}_count"] = (
                df1[f"{prefix_day}_{feature_name}_count"] - df1["is_nan_sum"]
            )
            df1[f"{prefix_day}_{feature_name}"] = df1[
                f"{prefix_day}_{feature_name}_sum"
            ] / (df1[f"{prefix_day}_{feature_name}_count"])

            df1.loc[
                df1[f"{prefix_day}_{feature_name}_count"] == 0,
                f"{prefix_day}_{feature_name}",
            ] = np.nan

        if self.add_count:
            df1[f"{prefix_day}_count"] = df1[f"{prefix_day}_{self.features[0]}_count"]
            df1 = df1.drop(
                columns=[
                    f"{prefix_day}_{feature_name}_count"
                    for feature_name in self.features
                ]
            )

        concat_df["__date_day"] = pd.to_datetime(concat_df["__date_day"])
        concat_df = concat_df.merge(
            df1[[c for c in df1.columns if c in self.features_out]].reset_index(),
            on=["__date_day", granularity_concat],
        )

        return concat_df

    def reset(self):
        self._df = None
        self._fitted_game_ids = []

    @property
    def features_out(self) -> list[str]:
        return self._features_out


class BinaryOutcomeRollingMeanTransformer(BaseLagGenerator):

    def __init__(
        self,
        features: list[str],
        window: int,
        binary_column: str,
        granularity: list[str] = None,
        prob_column: Optional[str] = None,
        min_periods: int = 1,
        add_opponent: bool = False,
        prefix: str = "rolling_mean_binary_",
    ):
        super().__init__(
            features=features,
            add_opponent=add_opponent,
            prefix=prefix,
            iterations=[],
            granularity=granularity,
        )
        self.window = window
        self.min_periods = min_periods
        self.binary_column = binary_column
        self.prob_column = prob_column
        for feature_name in self.features:
            feature1 = f"{self.prefix}{self.window}_{feature_name}_1"
            feature2 = f"{self.prefix}{self.window}_{feature_name}_0"
            self._features_out.append(f"{self.prefix}{self.window}_{feature_name}_1")
            self._features_out.append(f"{self.prefix}{self.window}_{feature_name}_0")
            self._entity_features.append(feature1)
            self._entity_features.append(feature2)

            if self.add_opponent:
                self._features_out.append(
                    f"{self.prefix}{self.window}_{feature_name}_1_opponent"
                )
                self._features_out.append(
                    f"{self.prefix}{self.window}_{feature_name}_0_opponent"
                )

        if self.prob_column:
            for feature_name in self.features:
                prob_feature = (
                    f"{self.prefix}{self.window}_{self.prob_column}_{feature_name}"
                )
                self._features_out.append(prob_feature)

        self._estimator_features_out = self._features_out.copy()

    def generate_historical(
        self, df: pd.DataFrame, column_names: ColumnNames
    ) -> pd.DataFrame:
        if df[self.binary_column].dtype in ("float64", "float32", "float16", "float"):
            df = df.assign(**{self.binary_column: df[self.binary_column].astype(int)})

        df = df.assign(is_future=0)
        self.column_names = column_names
        self.granularity = self.granularity or [self.column_names.player_id]
        validate_sorting(df=df, column_names=self.column_names)
        additional_cols_to_use = [self.binary_column] + (
            [self.prob_column] if self.prob_column else []
        )
        self._store_df(df, additional_cols_to_use=additional_cols_to_use)
        concat_df = self._generate_concat_df_with_feats(df)
        concat_df = self._add_weighted_prob(transformed_df=concat_df)
        transformed_df = self._create_transformed_df(df=df, concat_df=concat_df)
        if "is_future" in transformed_df.columns:
            transformed_df = transformed_df.drop(columns="is_future")
        return self._add_weighted_prob(transformed_df=transformed_df)

    def generate_future(self, df: pd.DataFrame) -> pd.DataFrame:
        if self._df is None:
            raise ValueError(
                "generate_historical needs to be called before generate_future"
            )

        if self.binary_column in df.columns:
            if df[self.binary_column].dtype in (
                "float64",
                "float32",
                "float16",
                "float",
            ):
                df = df.assign(
                    **{self.binary_column: df[self.binary_column].astype(int)}
                )
        df = df.assign(is_future=1)
        concat_df = self._generate_concat_df_with_feats(df=df)
        transformed_df = concat_df[
            concat_df[self.column_names.match_id].isin(
                df[self.column_names.match_id].astype("str").unique().tolist()
            )
        ]
        transformed_future = self._generate_future_feats(
            transformed_df=transformed_df,
            ori_df=df,
            known_future_features=self._get_known_future_features(),
        )
        if "is_future" in transformed_future.columns:
            transformed_future = transformed_future.drop(columns="is_future")
        return self._add_weighted_prob(transformed_df=transformed_future)

    def _get_known_future_features(self) -> list[str]:
        known_future_features = []
        if self.prob_column:
            for idx, feature_name in enumerate(self.features):
                weighted_prob_feat_name = (
                    f"{self.prefix}{self.window}_{self.prob_column}_{feature_name}"
                )
                known_future_features.append(weighted_prob_feat_name)

        return known_future_features

    def _add_weighted_prob(self, transformed_df: pd.DataFrame) -> pd.DataFrame:

        if self.prob_column:
            for idx, feature_name in enumerate(self.features):
                weighted_prob_feat_name = (
                    f"{self.prefix}{self.window}_{self.prob_column}_{feature_name}"
                )
                transformed_df[weighted_prob_feat_name] = transformed_df[
                    f"{self.prefix}{self.window}_{feature_name}_1"
                ] * transformed_df[self.prob_column] + transformed_df[
                    f"{self.prefix}{self.window}_{feature_name}_0"
                ] * (
                    1 - transformed_df[self.prob_column]
                )
        return transformed_df

    def _generate_concat_df_with_feats(self, df: pd.DataFrame) -> pd.DataFrame:

        additional_cols_to_use = [self.binary_column] + (
            [self.prob_column] if self.prob_column else []
        )
        concat_df = self._concat_df(df, additional_cols_to_use=additional_cols_to_use)
        aggregation = {
            **{f: "mean" for f in self.features},
            **{self.binary_column: "first"},
        }
        grouped = (
            concat_df.groupby(
                [
                    self.column_names.update_match_id,
                    *self.granularity,
                    self.column_names.start_date,
                    "is_future",
                ]
            )
            .agg(aggregation)
            .reset_index()
        )

        grouped = grouped.sort_values(
            by=[
                self.column_names.start_date,
                "is_future",
                self.column_names.update_match_id,
            ]
        )

        feats_added = []
        for feature in self.features:
            mask_result_1 = grouped[self.binary_column] == 1
            mask_result_0 = grouped[self.binary_column] == 0

            grouped["value_result_1"] = grouped[feature].where(mask_result_1)
            grouped["value_result_0"] = grouped[feature].where(mask_result_0)

            grouped[f"{self.prefix}{self.window}_{feature}_1"] = grouped.groupby(
                self.granularity
            )["value_result_1"].transform(
                lambda x: x.shift().rolling(window=self.window, min_periods=1).mean()
            )
            grouped[f"{self.prefix}{self.window}_{feature}_0"] = grouped.groupby(
                self.granularity
            )["value_result_0"].transform(
                lambda x: x.shift().rolling(window=self.window, min_periods=1).mean()
            )

            feats_added.append(f"{self.prefix}{self.window}_{feature}_1")
            feats_added.append(f"{self.prefix}{self.window}_{feature}_0")

        grouped["count_result_1"] = (
            grouped[grouped[self.binary_column] == 1]
            .groupby([*self.granularity])
            .cumcount()
            + 1
        )
        grouped["count_result_0"] = (
            grouped[grouped[self.binary_column] == 0]
            .groupby([*self.granularity])
            .cumcount()
            + 1
        )
        grouped[["count_result_0", "count_result_1"]] = grouped.groupby(
            self.granularity
        )[["count_result_0", "count_result_1"]].fillna(method="ffill")
        for feature in self.features:
            grouped.loc[
                grouped["count_result_1"] < self.min_periods,
                f"{self.prefix}{self.window}_{feature}_1",
            ] = np.nan
            grouped.loc[
                grouped["count_result_0"] < self.min_periods,
                f"{self.prefix}{self.window}_{feature}_0",
            ] = np.nan

        concat_df = concat_df.merge(
            grouped[
                [self.column_names.update_match_id, *self.granularity, *feats_added]
            ],
            on=[self.column_names.update_match_id, *self.granularity],
            how="left",
        )

        concat_df[feats_added] = concat_df.groupby(self.granularity)[
            feats_added
        ].fillna(method="ffill")
        return concat_df


class RollingMeanTransformerPolars(BaseLagGeneratorPolars):
    """
    Calculates the rolling mean for a list of features over a window of matches.
    Rolling Mean Values are also shifted by one match to avoid data leakage.

    Use .generate_historical() to generate rolling mean for historical data.
        The historical data is stored as instance-variables after calling .generate_historical()
    Use .generate_future() to generate rolling mean for future data after having called .generate_historical()
    """

    def __init__(
        self,
        features: list[str],
        window: int,
        granularity: Union[list[str], str] = None,
        add_opponent: bool = False,
        min_periods: int = 1,
        are_estimator_features=True,
        prefix: str = "rolling_mean_",
    ):
        """
        :param features:   Features to create rolling mean for
        :param window: Window size for rolling mean.
            If 10 will calculate rolling mean over the prior 10 observations
        :param min_periods: Minimum number of observations in window required to have a non-null result
        :param granularity: Columns to group by before rolling mean. E.g. player_id or [player_id, position].
             In the latter case it will get the rolling mean for each player_id and position combination.
             Defaults to player_id
        :param add_opponent: If True, will add new columns containing rolling mean for the opponent team.
        :param are_estimator_features: If True, the features will be added to the estimator features.
            If false, it makes it possible for the outer layer (Pipeline) to exclude these features from the estimator.
        :param prefix: Prefix for the new rolling mean columns
        """

        super().__init__(
            features=features,
            add_opponent=add_opponent,
            iterations=[window],
            prefix=prefix,
            granularity=granularity,
            are_estimator_features=are_estimator_features,
        )
        self.window = window
        self.min_periods = min_periods

    def generate_historical(
        self, df: Union[pd.DataFrame, pl.DataFrame], column_names: ColumnNames
    ) -> Union[pl.DataFrame, pd.DataFrame]:
        """
        Generates rolling mean for historical data
        Stored the historical data as instance-variables so it's possible to generate future data afterwards
        The calculation is done using Polars.
         However, Pandas Dataframe can be used as input and it will also output a pandas dataframe in that case.

        :param df: Historical data
        :param column_names: Column names
        """
        if isinstance(df, pd.DataFrame):
            ori_type = "pd"
            df = pl.DataFrame(df)
        else:
            ori_type = "pl"
        if (
            df.unique(
                subset=[
                    column_names.player_id,
                    column_names.team_id,
                    column_names.match_id,
                ]
            ).shape[0]
            != df.shape[0]
        ):
            raise ValueError(
                f"Duplicated rows in df. Df must be a unique combination of {column_names.player_id} and {column_names.update_match_id}"
            )

        df = df.with_columns(pl.lit(0).alias("is_future"))

        self.column_names = column_names
        self.granularity = self.granularity or [self.column_names.player_id]
        self._store_df(df)
        concat_df = self._generate_concat_df_with_feats(df)
        transformed_df = self._create_transformed_df(df=df, concat_df=concat_df)
        cn = self.column_names
        recasts_mapping = {}
        for c in [cn.player_id, cn.team_id, cn.match_id]:
            if transformed_df[c].dtype != df[c].dtype:
                recasts_mapping[c] = df[c].dtype
        transformed_df = transformed_df.cast(recasts_mapping)

        df = df.join(
            transformed_df.select(
                cn.player_id, cn.team_id, cn.match_id, *self.features_out
            ),
            on=[cn.player_id, cn.team_id, cn.match_id],
            how="left",
        )
        if "is_future" in df.columns:
            df = df.drop("is_future")
        if ori_type == "pd":
            return df.to_pandas()
        return df

    def generate_future(
        self, df: Union[pd.DataFrame, pl.DataFrame]
    ) -> Union[pl.DataFrame, pd.DataFrame]:
        """
        Generates rolling mean for future data
        Assumes that .generate_historical() has been called before
        The calculation is done using Polars.
         However, Pandas Dataframe can be used as input and it will also output a pandas dataframe in that case.

        Regardless of the scheduled data of the future match, all future matches are perceived as being played in the same date.
        That is to ensure that a team's 2nd future match has the same rolling means as the 1st future match.
        :param df: Future data
        """

        if isinstance(df, pd.DataFrame):
            ori_type = "pd"
            df = pl.DataFrame(df)
        else:
            ori_type = "pl"

        df = df.with_columns(pl.lit(1).alias("is_future"))
        concat_df = self._generate_concat_df_with_feats(df=df)
        unique_match_ids = df.select(
            pl.col(self.column_names.match_id).cast(pl.Utf8).unique()
        ).to_series()
        transformed_df = concat_df.filter(
            pl.col(self.column_names.match_id).cast(pl.Utf8).is_in(unique_match_ids)
        )
        transformed_df = self._generate_future_feats(
            transformed_df=transformed_df, ori_df=df
        )
        cn = self.column_names
        recasts_mapping = {}
        for c in [cn.player_id, cn.team_id, cn.match_id]:
            if transformed_df[c].dtype != df[c].dtype:
                recasts_mapping[c] = df[c].dtype
        transformed_df = transformed_df.cast(recasts_mapping)

        df = df.join(
            transformed_df.select(
                cn.player_id, cn.team_id, cn.match_id, *self.features_out
            ),
            on=[cn.player_id, cn.team_id, cn.match_id],
            how="left",
        )
        if "is_future" in df.columns:
            df = df.drop("is_future")
        if ori_type == "pd":
            return df.to_pandas()
        return df

    def _generate_concat_df_with_feats(self, df: pl.DataFrame) -> pl.DataFrame:

        if self._df is None:
            raise ValueError("fit_transform needs to be called before transform")

        concat_df = self._concat_df(df)
        if self.column_names.participation_weight:
            concat_df = concat_df.with_columns(
                [
                    (
                        concat_df[feature_name]
                        * concat_df[self.column_names.participation_weight]
                    ).alias(feature_name)
                    for feature_name in self.features
                ]
            )

        agg_dict = [
            pl.col(feature_name).mean().alias(feature_name)
            for feature_name in self.features
        ]
        agg_dict.append(
            pl.col(self.column_names.start_date)
            .first()
            .alias(self.column_names.start_date)
        )

        grp = concat_df.group_by(
            self.granularity + [self.column_names.update_match_id]
        ).agg(agg_dict)
        grp = grp.filter(pl.col(self.column_names.start_date).is_not_null())
        grp = grp.sort(
            [self.column_names.start_date, self.column_names.update_match_id]
        )

        rolling_means = [
            pl.col(feature_name)
            .shift()
            .rolling_mean(window_size=self.window, min_periods=self.min_periods)
            .over(self.granularity)
            .alias(f"{self.prefix}{self.window}_{feature_name}")
            for feature_name in self.features
        ]
        grp = grp.with_columns(rolling_means)

        selection_columns = (
            self.granularity
            + [self.column_names.update_match_id]
            + [f"{self.prefix}{self.window}_{feature}" for feature in self.features]
        )
        concat_df = concat_df.join(
            grp.select(selection_columns),
            on=self.granularity + [self.column_names.update_match_id],
            how="left",
        )

        concat_df = concat_df.sort(
            [
                self.column_names.start_date,
                self.column_names.match_id,
                self.column_names.team_id,
                self.column_names.player_id,
            ]
        )

        feats_added = [f for f in self.features_out if f in concat_df.columns]

        concat_df = concat_df.with_columns(
            [
                pl.col(f).forward_fill().over(self.granularity).alias(f)
                for f in feats_added
            ]
        )
        return concat_df

    @property
    def features_out(self) -> list[str]:
        return self._features_out



import copy
import logging
from dataclasses import dataclass, field
from typing import Optional, Union

import pandas as pd

from player_performance_ratings.ratings.performance_generator.performances_transformers import (
    SymmetricDistributionTransformer,
    MinMaxTransformer,
    PartialStandardScaler,
)


from player_performance_ratings.transformers.base_transformer import (
    BasePerformancesTransformer,
)


@dataclass
class ColumnWeight:
    name: str
    weight: float
    lower_is_better: bool = False

    def __post_init__(self):
        if self.weight < 0:
            raise ValueError("Weight must be positive")
        if self.weight > 1:
            raise ValueError("Weight must be less than 1")


@dataclass
class Performance:
    name: str = "performance"
    weights: list[ColumnWeight] = field(default_factory=list)


def auto_create_pre_performance_transformations(
    pre_transformers: list[BasePerformancesTransformer],
    performances: list[Performance],
) -> list[BasePerformancesTransformer]:
    """
    Creates a list of transformers that ensure the performance column is generated in a way that makes sense for the rating model.
    Ensures columns aren't too skewed, scales them to similar ranges, ensure values are between 0 and 1,
    and then weights them according to the column_weights.
    """

    not_transformed_features = []
    transformed_features = []

    for performance in performances:
        not_transformed_features += [
            p.name
            for p in performance.weights
            if p.name not in not_transformed_features
        ]

    distribution_transformer = SymmetricDistributionTransformer(
        features=not_transformed_features, prefix=""
    )
    pre_transformers.append(distribution_transformer)

    all_features = transformed_features + not_transformed_features

    pre_transformers.append(
        PartialStandardScaler(
            features=all_features, ratio=1, max_value=9999, target_mean=0, prefix=""
        )
    )
    pre_transformers.append(MinMaxTransformer(features=all_features))

    return pre_transformers


class PerformancesGenerator:

    def __init__(
        self,
        performances: Union[list[Performance], Performance],
        transformers: Optional[list[BasePerformancesTransformer]] = None,
        auto_transform_performance: bool = True,
    ):

        self.performances = (
            performances if isinstance(performances, list) else [performances]
        )
        self.auto_transform_performance = auto_transform_performance
        self.original_transformers = (
            [copy.deepcopy(p) for p in transformers] if transformers else []
        )
        self.transformers = transformers or []
        if self.auto_transform_performance:
            self.transformers = auto_create_pre_performance_transformations(
                pre_transformers=self.transformers, performances=self.performances
            )

    def generate(self, df):

        if self.transformers:
            for transformer in self.transformers:
                df = transformer.fit_transform(df)

        for performance in self.performances:
            if self.transformers:
                max_idx = len(self.transformers) - 1
                column_weighs_mapping = {
                    col_weight.name: self.transformers[max_idx].features_out[idx]
                    for idx, col_weight in enumerate(performance.weights)
                }
            else:
                column_weighs_mapping = None

            df[performance.name] = self._weight_columns(
                df=df,
                performance_column_name=performance.name,
                col_weights=performance.weights,
                column_weighs_mapping=column_weighs_mapping,
            )

            if df[performance.name].isnull().any():
                logging.error(
                    f"df[{performance.name}] contains nan values. Make sure all column_names used in column_weights are imputed beforehand"
                )
                raise ValueError("performance contains nan values")

        return df

    def _weight_columns(
        self,
        df: pd.DataFrame,
        performance_column_name: str,
        col_weights: list[ColumnWeight],
        column_weighs_mapping: dict[str, str],
    ) -> pd.DataFrame:
        df = df.copy()
        df[f"__{performance_column_name}"] = 0

        df["sum_cols_weights"] = 0
        for column_weight in col_weights:
            df[f"weight__{column_weight.name}"] = column_weight.weight
            df.loc[df[column_weight.name].isna(), f"weight__{column_weight.name}"] = 0
            df.loc[df[column_weight.name].isna(), column_weight.name] = 0
            df["sum_cols_weights"] = (
                df["sum_cols_weights"] + df[f"weight__{column_weight.name}"]
            )

        drop_cols = ["sum_cols_weights", f"__{performance_column_name}"]
        for column_weight in col_weights:
            df[f"weight__{column_weight.name}"] / df["sum_cols_weights"]
            drop_cols.append(f"weight__{column_weight.name}")

        sum_weight = sum([w.weight for w in col_weights])

        for column_weight in col_weights:

            if column_weighs_mapping:
                feature_name = column_weighs_mapping[column_weight.name]
            else:
                feature_name = column_weight.name

            if column_weight.lower_is_better:
                df[f"__{performance_column_name}"] += (
                    df[f"weight__{column_weight.name}"]
                    / sum_weight
                    * (1 - df[feature_name])
                )
            else:
                df[f"__{performance_column_name}"] += (
                    df[f"weight__{column_weight.name}"] / sum_weight * df[feature_name]
                )

        return df[f"__{performance_column_name}"].clip(0, 1)

    @property
    def features_out(self) -> list[str]:
        return [c.name for c in self.performances]



from .lag_generators import (
    LagTransformer,
    RollingMeanTransformer,
    RollingMeanDaysTransformer,
    BinaryOutcomeRollingMeanTransformer,
)
from .transformers import (
    RatioTeamPredictorTransformer,
    PredictorTransformer,
    NetOverPredictedPostTransformer,
    ModifierTransformer,
    ModifyOperation,
    Operation,
)



import copy
import inspect
import logging
from abc import abstractmethod, ABC
from typing import Optional, Match

import optuna
import pandas as pd
from optuna.samplers import TPESampler
from optuna.trial import BaseTrial
from player_performance_ratings.tuner.start_rating_optimizer import (
    StartLeagueRatingOptimizer,
)

from player_performance_ratings.cross_validator.cross_validator import CrossValidator
from player_performance_ratings.ratings.rating_calculators import MatchRatingGenerator

from player_performance_ratings.ratings.rating_calculators.start_rating_generator import (
    StartRatingGenerator,
)
from player_performance_ratings.ratings import UpdateRatingGenerator

from player_performance_ratings.ratings.rating_generator import RatingGenerator

from player_performance_ratings import PipelineFactory
from player_performance_ratings.tuner.utils import (
    ParameterSearchRange,
    add_params_from_search_range,
)

DEFAULT_TEAM_SEARCH_RANGES = [
    ParameterSearchRange(name="confidence_weight", type="uniform", low=0.6, high=0.95),
    ParameterSearchRange(
        name="confidence_days_ago_multiplier",
        type="uniform",
        low=0.02,
        high=0.12,
    ),
    ParameterSearchRange(
        name="confidence_max_days",
        type="uniform",
        low=40,
        high=200,
    ),
    ParameterSearchRange(
        name="confidence_max_sum",
        type="uniform",
        low=20,
        high=180,
    ),
    ParameterSearchRange(
        name="confidence_value_denom",
        type="uniform",
        low=20,
        high=180,
    ),
    ParameterSearchRange(
        name="rating_change_multiplier", type="uniform", low=25, high=140
    ),
    ParameterSearchRange(
        name="min_rating_change_multiplier_ratio",
        type="uniform",
        low=0.05,
        high=0.2,
    ),
    ParameterSearchRange(
        name="team_id_change_confidence_sum_decrease",
        type="uniform",
        low=0,
        high=10,
    ),
]

DEFAULT_START_RATING_SEARCH_RANGE = [
    ParameterSearchRange(
        name="league_quantile",
        type="uniform",
        low=0.12,
        high=0.5,
    ),
    ParameterSearchRange(
        name="min_count_for_percentiles",
        type="int",
        low=50,
        high=200,
    ),
    ParameterSearchRange(name="team_rating_subtract", type="int", low=0, high=200),
    ParameterSearchRange(name="team_weight", type="uniform", low=0, high=1),
]


class RatingGeneratorTuner(ABC):

    @abstractmethod
    def tune(
        self,
        df: pd.DataFrame,
        rating_idx: int,
        cross_validator: CrossValidator,
        pipeline_factory: PipelineFactory,
        matches: list[Match],
    ) -> RatingGenerator:
        pass


class UpdateRatingGeneratorTuner(RatingGeneratorTuner):

    def __init__(
        self,
        team_rating_search_ranges: Optional[list[ParameterSearchRange]] = None,
        team_rating_n_trials: int = 30,
        start_rating_search_ranges: Optional[list[ParameterSearchRange]] = None,
        start_rating_n_trials: int = 8,
        optimize_league_ratings: bool = False,
    ):
        self.team_rating_search_ranges = (
            team_rating_search_ranges or DEFAULT_TEAM_SEARCH_RANGES
        )
        self.start_rating_search_ranges = (
            start_rating_search_ranges or DEFAULT_START_RATING_SEARCH_RANGE
        )
        self.team_rating_n_trials = team_rating_n_trials
        self.start_rating_n_trials = start_rating_n_trials
        self.optimize_league_ratings = optimize_league_ratings

    def tune(
        self,
        df: pd.DataFrame,
        rating_idx: int,
        cross_validator: CrossValidator,
        pipeline_factory: PipelineFactory,
        matches: list[Match],
    ) -> UpdateRatingGenerator:

        if pipeline_factory.rating_generators:
            best_rating_generator = copy.deepcopy(
                pipeline_factory.rating_generators[rating_idx]
            )
        else:
            raise ValueError("rating_generators are not specified")
        # potential_rating_features = [v for k, v in RatingColumnNames.__dict__.items() if isinstance(v, str)]

        # best_rating_generator = UpdateRatingGenerator(
        # features_out=[f for f in pipeline_factory.predictor.features if f in potential_rating_features],
        #      features_out=pipeline_factory.predictor.features,
        #      column_names=pipeline_factory.rating_generators[rating_idx].column_names
        #  )

        if self.team_rating_n_trials > 0:
            logging.info("Tuning Team Rating")
            best_team_rating_generator = self._tune_team_rating(
                df=df,
                rating_generator=best_rating_generator,
                rating_index=rating_idx,
                cross_validator=cross_validator,
                matches=matches,
                pipeline_factory=pipeline_factory,
            )

            best_rating_generator.match_rating_generator = best_team_rating_generator

        if self.optimize_league_ratings:
            start_rating_optimizer = StartLeagueRatingOptimizer(
                pipeline_factory=pipeline_factory,
                cross_validator=cross_validator,
            )
            optimized_league_ratings = start_rating_optimizer.optimize(
                df=df,
                rating_model_idx=rating_idx,
                matches=matches,
                rating_generator=best_rating_generator,
                column_names=pipeline_factory.column_names,
            )
            best_rating_generator.match_rating_generator.start_rating_generator.league_ratings = (
                optimized_league_ratings
            )

        if self.start_rating_n_trials > 0:
            logging.info("Tuning Start Rating")

            best_start_rating = self._tune_start_rating(
                df=df,
                matches=matches,
                rating_generator=best_rating_generator,
                rating_index=rating_idx,
                cross_validator=cross_validator,
                match_predictor_factory=pipeline_factory,
            )
            best_rating_generator.match_rating_generator.start_rating_generator = (
                best_start_rating
            )

        return UpdateRatingGenerator(
            match_rating_generator=best_rating_generator.match_rating_generator,
            non_estimator_known_features_out=best_rating_generator.non_estimator_known_features_out,
            known_features_out=best_rating_generator._known_features_out,
            performance_column=best_rating_generator.performance_column,
        )

    def _tune_team_rating(
        self,
        df: pd.DataFrame,
        rating_generator: UpdateRatingGenerator,
        rating_index: int,
        matches: list[Match],
        cross_validator: CrossValidator,
        pipeline_factory: PipelineFactory,
    ) -> MatchRatingGenerator:

        def objective(trial: BaseTrial, df: pd.DataFrame) -> float:

            team_rating_generator_params = list(
                inspect.signature(
                    rating_generator.match_rating_generator.__class__.__init__
                ).parameters.keys()
            )[1:]

            params = {
                attr: getattr(rating_generator.match_rating_generator, attr)
                for attr in team_rating_generator_params
                if attr not in ("performance_predictor", "start_rating_generator")
            }
            params = add_params_from_search_range(
                params=params,
                trial=trial,
                parameter_search_range=self.team_rating_search_ranges,
            )

            performance_predictor = copy.deepcopy(
                rating_generator.match_rating_generator.performance_predictor
            )
            performance_predictor_params = list(
                inspect.signature(
                    performance_predictor.__class__.__init__
                ).parameters.keys()
            )[1:]

            for param in params.copy():
                if param in performance_predictor_params:
                    performance_predictor.__setattr__(param, params[param])
                    params.pop(param)

            team_rating_generator = MatchRatingGenerator(
                **params,
                performance_predictor=performance_predictor,
                start_rating_generator=copy.deepcopy(
                    rating_generator.match_rating_generator.start_rating_generator
                )
            )

            rating_g = copy.deepcopy(rating_generator)
            rating_g.match_rating_generator = team_rating_generator
            rating_generators = copy.deepcopy(pipeline_factory.rating_generators)

            rating_generators[rating_index] = rating_g
            pipeline = pipeline_factory.create(
                rating_generators=rating_generators,
            )
            return pipeline.cross_validate_score(
                df=df,
                matches=matches,
                cross_validator=cross_validator,
                create_performance=False,
            )

        direction = "minimize"
        study_name = "optuna_study"
        optuna_seed = 12
        sampler = TPESampler(seed=optuna_seed)
        study = optuna.create_study(
            direction=direction, study_name=study_name, sampler=sampler
        )
        callbacks = []
        study.optimize(
            lambda trial: objective(trial, df),
            n_trials=self.team_rating_n_trials,
            callbacks=callbacks,
        )

        best_params = study.best_params
        team_rating_generator_params = list(
            inspect.signature(
                rating_generator.match_rating_generator.__class__.__init__
            ).parameters.keys()
        )[1:]
        other_params = {
            attr: getattr(rating_generator.match_rating_generator, attr)
            for attr in team_rating_generator_params
            if attr not in ("performance_predictor", "start_rating_generator")
        }
        for param in other_params:
            if param not in best_params:
                best_params[param] = other_params[param]

        performance_predictor = (
            rating_generator.match_rating_generator.performance_predictor
        )
        performance_predictor_params = list(
            inspect.signature(
                performance_predictor.__class__.__init__
            ).parameters.keys()
        )[1:]

        for param in best_params.copy():
            if param in performance_predictor_params:
                performance_predictor.__setattr__(param, best_params[param])
                best_params.pop(param)
        return MatchRatingGenerator(
            **best_params,
            performance_predictor=performance_predictor,
            start_rating_generator=rating_generator.match_rating_generator.start_rating_generator
        )

    def _tune_start_rating(
        self,
        df: pd.DataFrame,
        rating_generator: UpdateRatingGenerator,
        rating_index: int,
        matches: list[Match],
        cross_validator: CrossValidator,
        match_predictor_factory: PipelineFactory,
    ):
        def objective(trial: BaseTrial, df: pd.DataFrame) -> float:
            start_rating_generator_params = list(
                inspect.signature(
                    rating_generator.match_rating_generator.start_rating_generator.__class__.__init__
                ).parameters.keys()
            )[1:]

            params = {
                attr: getattr(
                    rating_generator.match_rating_generator.start_rating_generator, attr
                )
                for attr in start_rating_generator_params
                if attr != "league_ratings"
            }

            params = add_params_from_search_range(
                params=params,
                trial=trial,
                parameter_search_range=self.start_rating_search_ranges,
            )

            league_ratings = copy.deepcopy(
                rating_generator.match_rating_generator.start_rating_generator.league_ratings
            )
            start_rating_generator = StartRatingGenerator(
                league_ratings=league_ratings, **params
            )
            rating_g = copy.deepcopy(rating_generator)
            rating_g.match_rating_generator.start_rating_generator = (
                start_rating_generator
            )
            if match_predictor_factory.rating_generators:
                rating_generators = copy.deepcopy(
                    match_predictor_factory.rating_generators
                )
                rating_generators[rating_index] = rating_g
            else:
                rating_generators = [rating_g]
            match_predictor = match_predictor_factory.create(
                rating_generators=rating_generators,
            )

            return match_predictor.cross_validate_score(
                df=df,
                matches=matches,
                create_performance=False,
                cross_validator=cross_validator,
            )

        direction = "minimize"
        study_name = "optuna_study"
        optuna_seed = 12
        sampler = TPESampler(seed=optuna_seed)
        study = optuna.create_study(
            direction=direction, study_name=study_name, sampler=sampler
        )
        callbacks = []
        study.optimize(
            lambda trial: objective(trial, df),
            n_trials=self.start_rating_n_trials,
            callbacks=callbacks,
        )
        start_rating_generator_params = list(
            inspect.signature(
                rating_generator.match_rating_generator.start_rating_generator.__class__.__init__
            ).parameters.keys()
        )[1:]

        other_params = {
            attr: getattr(
                rating_generator.match_rating_generator.start_rating_generator, attr
            )
            for attr in start_rating_generator_params
        }

        best_params = study.best_params
        for param in other_params:
            if param not in best_params:
                best_params[param] = other_params[param]

        return StartRatingGenerator(**best_params)



import math
import time
from abc import ABC, abstractmethod
from typing import Optional

import numpy as np
import pandas as pd
from lightgbm import LGBMRegressor

from player_performance_ratings.data_structures import (
    PreMatchPlayerRating,
    PreMatchTeamRating,
)

MATCH_CONTRIBUTION_TO_SUM_VALUE = 1
MODIFIED_RATING_CHANGE_CONSTANT = 1
CERTAIN_SUM = "certain_sum"


def sigmoid_subtract_half_and_multiply2(value: float, x: float) -> float:
    return (1 / (1 + math.exp(-value / x)) - 0.5) * 2


class PerformancePredictor(ABC):

    @abstractmethod
    def reset(self):
        pass

    @abstractmethod
    def predict_performance(
        self,
        player_rating: PreMatchPlayerRating,
        opponent_team_rating: PreMatchTeamRating,
        team_rating: PreMatchTeamRating,
    ) -> float:
        pass


class RatingMeanCustomOpponentRatingPerformancePredictor(PerformancePredictor):

    def __init__(
        self,
        opponent_rating_column_name: str,
        opponent_rating_std: float,
        opponent_rating_mean: float,
        player_rating_expected_mean: float = 1000,
        min_count: int = 2500,
        coef: float = 0.001757,
    ):
        self.opponent_rating_column_name = opponent_rating_column_name
        self.opponent_rating_std = opponent_rating_std
        self.opponent_rating_mean = opponent_rating_mean
        self.player_rating_expected_mean = player_rating_expected_mean
        self.min_count = min_count
        self.coef = coef
        self._player_ratings = []
        self._backup_performance_predictor = RatingNonOpponentPerformancePredictor()
        self._sum_rating = 0
        self._rating_count = 0
        self._actual_player_rating_std = 0
        self._actual_player_rating_mean = 0

    def reset(self):
        self._player_ratings = []
        self._sum_rating = 0
        self._rating_count = 0
        self._actual_player_rating_std = 0
        self._actual_player_rating_mean = 0
        self._backup_performance_predictor.reset()

    def predict_performance(
        self,
        player_rating: PreMatchPlayerRating,
        opponent_team_rating: PreMatchTeamRating,
        team_rating: PreMatchTeamRating,
    ) -> float:

        opponent_team_rating_value = sum(
            [
                opponent_player.rating_value
                for opponent_player in opponent_team_rating.players
            ]
        ) / len(opponent_team_rating.players)

        if len(self._player_ratings) == 0:
            actual_player_rating_mean = 0
            weight_actual_data = 0
        else:
            weight_actual_data = min(1, len(self._player_ratings) / 4000)

            if (
                len(self._player_ratings) % 200 == 0
                and len(self._player_ratings) >= self.min_count
                or self._actual_player_rating_std == 0
                and len(self._player_ratings) >= self.min_count
            ):
                self._actual_player_rating_std = np.array(self._player_ratings).std()
                self._actual_player_rating_mean = sum(self._player_ratings) / len(
                    self._player_ratings
                )

        player_rating_mean = (
            (1 - weight_actual_data) * self.player_rating_expected_mean
            + weight_actual_data * self._actual_player_rating_mean
        )

        player_rating_std = (
            (1 - weight_actual_data) * self.opponent_rating_std
            + weight_actual_data * self._actual_player_rating_std
        )
        opponent_rating_adjusted = (
            (
                opponent_team_rating_value
                - self.opponent_rating_mean
                + player_rating_mean
            )
            / self.opponent_rating_std
            * player_rating_std
        )

        if len(self._player_ratings) < self.min_count:
            predicted_performance = (
                self._backup_performance_predictor.predict_performance(
                    player_rating=player_rating,
                    opponent_team_rating=opponent_team_rating,
                    team_rating=team_rating,
                )
            )

        else:
            historical_average_rating = self._sum_rating / self._rating_count
            net_mean_rating_over_historical_average = (
                player_rating.rating_value * 0.5
                + opponent_rating_adjusted * 0.5
                - historical_average_rating
            )
            value = self.coef * net_mean_rating_over_historical_average
            predicted_performance = (math.exp(value)) / (1 + math.exp(value))

        self._sum_rating += (
            player_rating.rating_value * 0.5 + opponent_rating_adjusted * 0.5
        )
        self._rating_count += 1

        self._player_ratings.append(player_rating.rating_value)
        # print(predicted_performance, self._backup_performance_predictor.predict_performance(player_rating=player_rating,
        #                                                                                      opponent_team_rating=opponent_team_rating,
        #                                                                                     team_rating=team_rating))
        return predicted_performance


class RatingNonOpponentPerformancePredictor(PerformancePredictor):

    def __init__(
        self,
        coef: float = 0.0015,
        last_sample_count: int = 1500,
        min_count_for_historical_average: int = 200,
        historical_average_value_default: float = 1000,
    ):
        self.coef = coef
        self.last_sample_count = last_sample_count
        self.min_count_for_historical_average = min_count_for_historical_average
        self.historical_average_value_default = historical_average_value_default
        if self.min_count_for_historical_average < 1:
            raise ValueError("min_count_for_historical_average must be positive")
        self._prev_entries_ratings = []

    def reset(self):
        self._prev_entries_ratings = []

    def predict_performance(
        self,
        player_rating: PreMatchPlayerRating,
        opponent_team_rating: PreMatchTeamRating,
        team_rating: PreMatchTeamRating,
    ) -> float:
        start_index = max(0, len(self._prev_entries_ratings) - self.last_sample_count)
        recent_prev_entries_ratings = self._prev_entries_ratings[start_index:]
        if len(recent_prev_entries_ratings) > self.min_count_for_historical_average:
            historical_average_rating = sum(recent_prev_entries_ratings) / len(
                recent_prev_entries_ratings
            )
        else:
            historical_average_rating = self.historical_average_value_default
        net_mean_rating_over_historical_average = (
            player_rating.rating_value - historical_average_rating
        )

        value = self.coef * net_mean_rating_over_historical_average
        prediction = (math.exp(value)) / (1 + math.exp(value))
        self._prev_entries_ratings.append(player_rating.rating_value)

        return prediction


class RatingDifferencePerformancePredictor(PerformancePredictor):

    # TODO: Performance prediction based on team-players sharing time with.
    def __init__(
        self,
        rating_diff_coef: float = 0.005757,
        rating_diff_team_from_entity_coef: float = 0.0,
        team_rating_diff_coef: float = 0.0,
        max_predict_value: float = 1,
        participation_weight_coef: Optional[float] = None,
    ):
        self.rating_diff_coef = rating_diff_coef
        self.rating_diff_team_from_entity_coef = rating_diff_team_from_entity_coef
        self.team_rating_diff_coef = team_rating_diff_coef
        self.max_predict_value = max_predict_value
        self.participation_weight_coef = participation_weight_coef

    def reset(self):
        pass

    def predict_performance(
        self,
        player_rating: PreMatchPlayerRating,
        opponent_team_rating: PreMatchTeamRating,
        team_rating: PreMatchTeamRating,
    ) -> float:

        if player_rating.match_performance.team_players_playing_time and isinstance(
            player_rating.match_performance.team_players_playing_time, dict
        ):
            weight_team_rating = 0
            sum_playing_time = 0
            for team_player in team_rating.players:
                if (
                    str(team_player.id)
                    in player_rating.match_performance.team_players_playing_time
                    or str(team_player.id)
                    in player_rating.match_performance.team_players_playing_time
                ):
                    weight_team_rating += (
                        team_player.rating_value
                        * player_rating.match_performance.team_players_playing_time[
                            str(team_player.id)
                        ]
                    )
                    sum_playing_time += (
                        player_rating.match_performance.team_players_playing_time[
                            str(team_player.id)
                        ]
                    )

            if sum_playing_time > 0:
                team_rating_value = weight_team_rating / sum_playing_time
            else:
                team_rating_value = team_rating.rating_value

        else:
            team_rating_value = team_rating.rating_value

        if (
            player_rating.match_performance.opponent_players_playing_time
            and isinstance(
                player_rating.match_performance.team_players_playing_time, dict
            )
        ):
            weight_opp_rating = 0
            sum_playing_time = 0
            for opp_player in opponent_team_rating.players:
                if (
                    str(opp_player.id)
                    in player_rating.match_performance.opponent_players_playing_time
                ):
                    weight_opp_rating += (
                        opp_player.rating_value
                        * player_rating.match_performance.opponent_players_playing_time[
                            str(opp_player.id)
                        ]
                    )
                    sum_playing_time += (
                        player_rating.match_performance.opponent_players_playing_time[
                            str(opp_player.id)
                        ]
                    )
            if sum_playing_time > 0:
                opp_rating_value = weight_opp_rating / sum_playing_time
            else:
                opp_rating_value = opponent_team_rating.rating_value

        else:
            opp_rating_value = opponent_team_rating.rating_value

        rating_difference = player_rating.rating_value - opp_rating_value

        rating_diff_team_from_entity = team_rating_value - player_rating.rating_value
        team_rating_diff = team_rating_value - opp_rating_value

        value = (
            self.rating_diff_coef * rating_difference
            + self.rating_diff_team_from_entity_coef * rating_diff_team_from_entity
            + team_rating_diff * self.team_rating_diff_coef
        )

        prediction = (math.exp(value)) / (1 + math.exp(value))

        if prediction > self.max_predict_value:
            return self.max_predict_value
        elif prediction < (1 - self.max_predict_value):
            return 1 - self.max_predict_value

        return prediction


class RatingMeanPerformancePredictor(PerformancePredictor):

    def __init__(
        self,
        coef: float = 0.005757,
        max_predict_value: float = 1,
        last_sample_count: int = 1500,
    ):
        self.coef = coef
        self.max_predict_value = max_predict_value
        self.last_sample_count = last_sample_count
        self.sum_ratings = []
        self.sum_rating = 0
        self.rating_count = 0

    def reset(self):
        self.sum_ratings = []
        self.sum_rating = 0
        self.rating_count = 0

    def predict_performance(
        self,
        player_rating: PreMatchPlayerRating,
        opponent_team_rating: PreMatchTeamRating,
        team_rating: PreMatchTeamRating,
    ) -> float:

        self.sum_ratings.append(player_rating.rating_value)
        self.rating_count += 1
        self.sum_rating += player_rating.rating_value
        start_index = max(0, len(self.sum_ratings) - self.last_sample_count)
        self.sum_ratings = self.sum_ratings[start_index:]
        #  average_rating = sum(self.sum_ratings) / len(self.sum_ratings)
        historical_average_rating = self.sum_rating / self.rating_count
        net_mean_rating_over_historical_average = (
            player_rating.rating_value * 0.5
            + opponent_team_rating.rating_value * 0.5
            - historical_average_rating
        )

        value = self.coef * net_mean_rating_over_historical_average
        prediction = (math.exp(value)) / (1 + math.exp(value))
        if prediction > self.max_predict_value:
            return self.max_predict_value
        elif prediction < (1 - self.max_predict_value):
            return 1 - self.max_predict_value
        return prediction



from unittest import mock

import pandas as pd
from player_performance_ratings.cross_validator import MatchKFoldCrossValidator

from player_performance_ratings.predictor import Predictor
from sklearn.linear_model import LinearRegression

from player_performance_ratings.ratings.rating_calculators import MatchRatingGenerator
from player_performance_ratings.ratings import (
    RatingKnownFeatures,
    UpdateRatingGenerator,
    RatingHistoricalFeatures,
)
from player_performance_ratings.ratings.performance_generator import (
    Performance,
    ColumnWeight,
    PerformancesGenerator,
)

from player_performance_ratings.transformers import (
    LagTransformer,
    RatioTeamPredictorTransformer,
    PredictorTransformer,
    RollingMeanTransformer,
)

from player_performance_ratings import ColumnNames, Pipeline
from player_performance_ratings.transformers.lag_generators import (
    RollingMeanTransformerPolars,
)


def test_pipline_mix_pandas_polars_lags():
    df = pd.DataFrame(
        {
            "game_id": [1, 1, 2, 2, 3, 3, 4, 4],
            "player_id": [1, 2, 1, 2, 1, 2, 1, 2],
            "team_id": [1, 2, 1, 2, 1, 2, 1, 2],
            "start_date": [
                pd.to_datetime("2023-01-01"),
                pd.to_datetime("2023-01-01"),
                pd.to_datetime("2023-01-02"),
                pd.to_datetime("2023-01-02"),
                pd.to_datetime("2023-01-03"),
                pd.to_datetime("2023-01-03"),
                pd.to_datetime("2023-01-04"),
                pd.to_datetime("2023-01-05"),
            ],
            "deaths": [1, 1, 1, 2, 2, 2, 3, 2],
            "kills": [0.2, 0.3, 0.4, 0.5, 2, 0.2, 2, 1],
            "__target": [1, 0, 1, 0, 1, 0, 1, 0],
        }
    )
    lag_generators = [
        RollingMeanTransformer(
            features=["kills", "deaths"],
            window=1,
            granularity=["player_id"],
            prefix="rolling_mean_",
        ),
        RollingMeanTransformerPolars(
            features=["kills", "deaths"],
            window=2,
            granularity=["player_id"],
            prefix="rolling_mean_polars_",
        ),
        RollingMeanTransformer(
            features=["kills", "deaths"],
            window=3,
            granularity=["player_id"],
            prefix="rolling_mean_",
        ),
    ]

    rating_generator = UpdateRatingGenerator(
        known_features_out=[RatingKnownFeatures.RATING_DIFFERENCE_PROJECTED],
        historical_features_out=[RatingHistoricalFeatures.TEAM_RATING],
        performance_column="kills",
    )

    pipeline = Pipeline(
        column_names=ColumnNames(
            match_id="game_id",
            team_id="team_id",
            player_id="player_id",
            start_date="start_date",
        ),
        lag_generators=lag_generators,
        performances_generator=PerformancesGenerator(
            performances=Performance(name="kills")
        ),
        predictor=Predictor(estimator=LinearRegression()),
        rating_generators=rating_generator,
    )

    df_with_predict_and_features = pipeline.train_predict(df=df, return_features=True)
    pipeline.cross_validate_predict(
        df=df,
        cross_validator=MatchKFoldCrossValidator(
            match_id_column_name="game_id", n_splits=1, date_column_name="start_date"
        ),
    )
    df_with_future = pipeline.future_predict(df=df, return_features=True)
    assert df_with_predict_and_features.shape[0] == df_with_future.shape[0]
    assert (
        RatingKnownFeatures.RATING_DIFFERENCE_PROJECTED
        in df_with_predict_and_features.columns
    )
    assert RatingHistoricalFeatures.TEAM_RATING in df_with_predict_and_features.columns


def test_pipeline_constructor():
    lag_generators = [
        RollingMeanTransformer(
            features=["kills", "deaths"],
            window=1,
            granularity=["player_id"],
            prefix="rolling_mean_",
        ),
    ]

    post_lag_transformers = [
        PredictorTransformer(predictor=Predictor(estimator=LinearRegression()))
    ]
    rating_generator = UpdateRatingGenerator(
        known_features_out=[RatingKnownFeatures.RATING_DIFFERENCE_PROJECTED]
    )

    pipeline = Pipeline(
        column_names=ColumnNames(
            match_id="game_id",
            team_id="team_id",
            player_id="player_id",
            start_date="start_date",
        ),
        lag_generators=lag_generators,
        post_lag_transformers=post_lag_transformers,
        predictor=Predictor(estimator=LinearRegression(), estimator_features=["kills"]),
        rating_generators=rating_generator,
    )

    expected_estimator_features = (
        ["kills"]
        + [l.features_out for l in lag_generators][0]
        + [p.predictor.pred_column for p in post_lag_transformers]
        + rating_generator.features_out
    )
    assert pipeline._estimator_features.sort() == expected_estimator_features.sort()

    # asserts estimator_features gets added to the post_transformer that contains a predictor
    assert (
        post_lag_transformers[0].features.sort()
        == [
            ["kills"]
            + [l.features_out for l in lag_generators][0]
            + rating_generator.features_out
        ][0].sort()
    )


def test_match_predictor_auto_pre_transformers():
    df = pd.DataFrame(
        {
            "game_id": [1, 1, 2, 2, 3, 3],
            "player_id": [1, 2, 3, 1, 2, 3],
            "team_id": [1, 2, 1, 2, 1, 3],
            "start_date": [
                pd.to_datetime("2023-01-01"),
                pd.to_datetime("2023-01-01"),
                pd.to_datetime("2023-01-02"),
                pd.to_datetime("2023-01-02"),
                pd.to_datetime("2023-01-03"),
                pd.to_datetime("2023-01-03"),
            ],
            "deaths": [1, 1, 1, 2, 2, 2],
            "kills": [0.2, 0.3, 0.4, 0.5, 2, 0.2],
            "__target": [1, 0, 1, 0, 1, 0],
        }
    )

    column_weights = [
        ColumnWeight(name="kills", weight=0.6),
        ColumnWeight(name="deaths", weight=0.4, lower_is_better=True),
    ]

    expected_df = df.copy()
    expected_df["prediction"] = [0.5, 0.5, 0.5, 0.5, 0.5, 0.5]

    predictor_mock = mock.Mock()
    predictor_mock.target = "__target"
    predictor_mock.columns_added = ["prediction"]
    predictor_mock._estimator_features = [
        RatingKnownFeatures.RATING_DIFFERENCE_PROJECTED
    ]

    predictor_mock.add_prediction.return_value = expected_df
    rating_generators = UpdateRatingGenerator(
        known_features_out=[RatingKnownFeatures.RATING_DIFFERENCE_PROJECTED],
        performance_column="weighted_performance",
    )

    pipeline = Pipeline(
        predictor=predictor_mock,
        rating_generators=rating_generators,
        performances_generator=PerformancesGenerator(
            performances=Performance(
                name="weighted_performance", weights=column_weights
            )
        ),
        column_names=ColumnNames(
            match_id="game_id",
            team_id="team_id",
            player_id="player_id",
            start_date="start_date",
        ),
    )

    new_df = pipeline.train_predict(df=df)

    pd.testing.assert_frame_equal(new_df, expected_df, check_like=True)

    assert len(pipeline.performances_generator.transformers) > 0


def test_match_predictor_multiple_rating_generators_same_performance():
    df = pd.DataFrame(
        {
            "game_id": [1, 1, 2, 2, 3, 3],
            "player_id": [1, 2, 3, 1, 2, 3],
            "team_id": [1, 2, 1, 2, 1, 3],
            "start_date": [
                pd.to_datetime("2023-01-01"),
                pd.to_datetime("2023-01-01"),
                pd.to_datetime("2023-01-02"),
                pd.to_datetime("2023-01-02"),
                pd.to_datetime("2023-01-03"),
                pd.to_datetime("2023-01-03"),
            ],
            "performance": [0.2, 0.8, 0.4, 0.6, 1, 0],
            "__target": [1, 0, 1, 0, 1, 0],
        }
    )

    column_names1 = ColumnNames(
        match_id="game_id",
        team_id="team_id",
        player_id="player_id",
        start_date="start_date",
    )

    df = df.sort_values(
        by=[
            column_names1.start_date,
            column_names1.match_id,
            column_names1.team_id,
            column_names1.player_id,
        ]
    )

    expected_df = df.copy()
    expected_df["prediction"] = [0.5, 0.5, 0.5, 0.5, 0.5, 0.5]

    predictor_mock = mock.Mock()
    predictor_mock.target = "__target"
    predictor_mock.add_prediction.return_value = expected_df
    predictor_mock.columns_added = ["prediction"]
    predictor_mock._estimator_features = [
        RatingKnownFeatures.RATING_DIFFERENCE_PROJECTED
    ]

    match_predictor = Pipeline(
        rating_generators=[
            UpdateRatingGenerator(
                known_features_out=[RatingKnownFeatures.RATING_DIFFERENCE_PROJECTED],
                prefix="rating_1",
            ),
            UpdateRatingGenerator(
                match_rating_generator=MatchRatingGenerator(
                    rating_change_multiplier=20
                ),
                known_features_out=[RatingKnownFeatures.RATING_DIFFERENCE_PROJECTED],
                prefix="rating_2",
            ),
        ],
        lag_generators=[],
        predictor=predictor_mock,
        column_names=column_names1,
    )

    new_df = match_predictor.train_predict(df=df)
    pd.testing.assert_frame_equal(new_df, expected_df, check_like=True)

    col_names_predictor_train = predictor_mock.train.call_args[1]["df"].columns.tolist()

    col_names_predictor_add = predictor_mock.add_prediction.call_args[1][
        "df"
    ].columns.tolist()

    assert (
        match_predictor.rating_generators[0].features_out[0] in col_names_predictor_add
    )
    assert (
        match_predictor.rating_generators[0].features_out[0]
        in col_names_predictor_train
    )

    assert (
        match_predictor.rating_generators[1].features_out[0] in col_names_predictor_add
    )
    assert (
        match_predictor.rating_generators[1].features_out[0]
        in col_names_predictor_train
    )


def test_match_predictor_0_rating_generators():
    """
    Post rating transformers are used, but no rating model. the features from transformers should be used to train model and add prediction
    """

    df = pd.DataFrame(
        {
            "game_id": [1, 1, 2, 2, 3, 3],
            "player_id": [1, 2, 3, 1, 2, 3],
            "team_id": [1, 2, 1, 2, 1, 3],
            "start_date": [1, 1, 2, 2, 3, 3],
            "deaths": [1, 1, 1, 2, 2, 2],
            "kills": [0.2, 0.3, 0.4, 0.5, 2, 0.2],
            "__target": [1, 0, 1, 0, 1, 0],
        }
    )

    expected_df = df.copy()
    expected_df["prediction"] = [0.5, 0.5, 0.5, 0.5, 0.5, 0.5]

    predictor_mock = mock.Mock()
    predictor_mock.target = "__target"
    predictor_mock.columns_added = ["prediction"]
    predictor_mock.add_prediction.return_value = expected_df
    predictor_mock._estimator_features = [
        RatingKnownFeatures.RATING_DIFFERENCE_PROJECTED
    ]

    column_names = ColumnNames(
        match_id="game_id",
        team_id="team_id",
        player_id="player_id",
        start_date="start_date",
    )

    lag_transformer = LagTransformer(
        features=["kills", "deaths"],
        lag_length=1,
        granularity=["player_id"],
        prefix="lag_",
    )

    pipeline = Pipeline(
        rating_generators=[],
        lag_generators=[lag_transformer],
        predictor=predictor_mock,
        column_names=column_names,
    )

    new_df = pipeline.train_predict(df=df)

    pd.testing.assert_frame_equal(new_df, expected_df, check_like=True)

    col_names_predictor_train = predictor_mock.train.call_args[1]["df"].columns.tolist()
    assert any(
        lag_transformer.prefix in element for element in col_names_predictor_train
    )

    col_names_predictor_add = predictor_mock.add_prediction.call_args[1][
        "df"
    ].columns.tolist()
    assert any(lag_transformer.prefix in element for element in col_names_predictor_add)


def test_match_predictor_generate_and_predict():
    historical_df = pd.DataFrame(
        {
            "game_id": [1, 1, 2, 2, 3, 3],
            "player_id": [1, 2, 3, 1, 2, 3],
            "team_id": [1, 2, 1, 2, 1, 3],
            "start_date": [
                pd.to_datetime("2023-01-01"),
                pd.to_datetime("2023-01-01"),
                pd.to_datetime("2023-01-02"),
                pd.to_datetime("2023-01-02"),
                pd.to_datetime("2023-01-03"),
                pd.to_datetime("2023-01-03"),
            ],
            "deaths": [1, 1, 1, 2, 2, 2],
            "kills": [0.2, 0.3, 0.4, 0.5, 2, 0.2],
            "__target": [1, 0, 1, 0, 1, 0],
        }
    )

    future_df = pd.DataFrame(
        {
            "game_id": [4, 4, 5, 5],
            "player_id": [1, 2, 1, 3],
            "team_id": [1, 3, 1, 3],
            "start_date": [
                pd.to_datetime("2023-01-04"),
                pd.to_datetime("2023-01-04"),
                pd.to_datetime("2023-01-05"),
                pd.to_datetime("2023-01-05"),
            ],
        }
    )

    expected_future_df = future_df.copy()
    expected_future_df["prediction"] = [0.5, 0.5, 0.5, 0.5]

    column_weights = [
        ColumnWeight(name="kills", weight=0.6),
        ColumnWeight(name="deaths", weight=0.4, lower_is_better=True),
    ]

    historical_df_mock_return_with_prediction = historical_df.copy()
    historical_df_mock_return_with_prediction["prediction"] = [
        0.5,
        0.5,
        0.5,
        0.5,
        0.5,
        0.5,
    ]

    predictor_mock = mock.Mock()
    predictor_mock.target = "__target"
    predictor_mock.columns_added = ["prediction"]
    predictor_mock.add_prediction.side_effect = [
        historical_df_mock_return_with_prediction,
        expected_future_df,
    ]
    predictor_mock._estimator_features = [
        RatingKnownFeatures.RATING_DIFFERENCE_PROJECTED
    ]

    column_names = ColumnNames(
        match_id="game_id",
        team_id="team_id",
        player_id="player_id",
        start_date="start_date",
    )
    rating_generator = UpdateRatingGenerator(
        known_features_out=[RatingKnownFeatures.RATING_DIFFERENCE_PROJECTED]
    )

    pipeline = Pipeline(
        performances_generator=PerformancesGenerator(
            Performance(weights=column_weights)
        ),
        predictor=predictor_mock,
        rating_generators=rating_generator,
        column_names=column_names,
    )

    _ = pipeline.train_predict(df=historical_df)
    new_df = pipeline.future_predict(future_df)

    pd.testing.assert_frame_equal(new_df, expected_future_df, check_like=True)

    assert len(pipeline.performances_generator.transformers) > 0


def test_train_predict_cross_validate():
    historical_df = pd.DataFrame(
        {
            "game_id": [1, 1, 2, 2, 3, 3, 4, 4, 5, 5],
            "player_id": [1, 2, 1, 2, 1, 2, 1, 2, 1, 2],
            "team_id": [1, 2, 1, 2, 1, 2, 1, 2, 1, 2],
            "start_date": [
                pd.to_datetime("2023-01-01"),
                pd.to_datetime("2023-01-01"),
                pd.to_datetime("2023-01-02"),
                pd.to_datetime("2023-01-02"),
                pd.to_datetime("2023-01-03"),
                pd.to_datetime("2023-01-03"),
                pd.to_datetime("2023-01-03"),
                pd.to_datetime("2023-01-03"),
                pd.to_datetime("2023-01-04"),
                pd.to_datetime("2023-01-04"),
            ],
            "deaths": [1, 1, 1, 2, 2, 2, 1, 3, 2, 5],
            "kills": [0.2, 0.3, 0.4, 0.5, 2, 0.2, 1, 4, 3, 5],
            "__target": [1, 0, 1, 0, 1, 0, 1, 0, 1, 0],
        }
    )

    column_weights = [
        ColumnWeight(name="kills", weight=0.6),
        ColumnWeight(name="deaths", weight=0.4, lower_is_better=True),
    ]
    predictor = Predictor(estimator=LinearRegression())

    column_names = ColumnNames(
        match_id="game_id",
        team_id="team_id",
        player_id="player_id",
        start_date="start_date",
    )
    rating_generator = UpdateRatingGenerator(
        known_features_out=[RatingKnownFeatures.RATING_DIFFERENCE_PROJECTED]
    )

    pipeline = Pipeline(
        performances_generator=PerformancesGenerator(
            Performance(weights=column_weights)
        ),
        predictor=predictor,
        rating_generators=rating_generator,
        lag_generators=[
            LagTransformer(
                features=["kills", "deaths"], lag_length=1, granularity=["player_id"]
            )
        ],
        column_names=column_names,
    )

    cross_validated_df = pipeline.train_predict(
        df=historical_df, cross_validate_predict=True
    )


def test_cross_validate_is_equal_to_predict_future():
    df = pd.DataFrame(
        {
            "game_id": [1, 1, 2, 2, 3, 3, 4, 4, 5, 5],
            "player_id": [1, 2, 1, 2, 1, 2, 1, 2, 1, 2],
            "team_id": [1, 2, 1, 2, 1, 2, 1, 2, 1, 2],
            "start_date": [
                pd.to_datetime("2023-01-01"),
                pd.to_datetime("2023-01-01"),
                pd.to_datetime("2023-01-02"),
                pd.to_datetime("2023-01-02"),
                pd.to_datetime("2023-01-03"),
                pd.to_datetime("2023-01-03"),
                pd.to_datetime("2023-01-03"),
                pd.to_datetime("2023-01-03"),
                pd.to_datetime("2023-01-04"),
                pd.to_datetime("2023-01-04"),
            ],
            "deaths": [0.5, 1, 0.7, 2, 0.5, 0.7, 0.2, 2.1, 0.8, 1],
            "kills": [0.2, 0.3, 0.2, 0.3, 0.2, 0.3, 0.2, 0.3, 0.2, 0.3],
            "__target": [1, 0, 0.6, 0.3, 0.8, 0.2, 0.4, 0.1, 1, 0],
        }
    )

    column_weights = [ColumnWeight(name="kills", weight=1)]
    predictor = Predictor(estimator=LinearRegression())

    column_names = ColumnNames(
        match_id="game_id",
        team_id="team_id",
        player_id="player_id",
        start_date="start_date",
    )
    rating_generator = UpdateRatingGenerator(
        known_features_out=[RatingKnownFeatures.RATING_DIFFERENCE_PROJECTED]
    )

    pipeline = Pipeline(
        performances_generator=PerformancesGenerator(
            Performance(weights=column_weights)
        ),
        predictor=predictor,
        rating_generators=rating_generator,
        lag_generators=[
            LagTransformer(
                features=["kills", "deaths"], lag_length=1, granularity=["player_id"]
            )
        ],
        column_names=column_names,
    )

    cross_validator = MatchKFoldCrossValidator(
        match_id_column_name=column_names.match_id,
        n_splits=1,
        min_validation_date="2023-01-04",
        date_column_name=column_names.start_date,
    )
    cross_validated_df = pipeline.cross_validate_predict(
        df=df, cross_validator=cross_validator, add_train_prediction=True
    )

    historical_df = df[df[column_names.start_date] < pd.to_datetime("2023-01-04")]
    future_df = df[df[column_names.start_date] >= pd.to_datetime("2023-01-04")]
    historical_df_predictions = pipeline.train_predict(df=historical_df)
    future_predict = pipeline.future_predict(df=future_df).reset_index(drop=True)
    future_cv_df = cross_validated_df[
        cross_validated_df[column_names.start_date] >= pd.to_datetime("2023-01-04")
    ].reset_index(drop=True)
    past_cv_df = cross_validated_df[
        cross_validated_df[column_names.start_date] < pd.to_datetime("2023-01-04")
    ].reset_index(drop=True)
    pd.testing.assert_frame_equal(
        future_cv_df, future_predict, check_like=True, check_dtype=False
    )
    pd.testing.assert_frame_equal(
        past_cv_df, historical_df_predictions, check_like=True, check_dtype=False
    )


def test_train_predict_cross_validate_is_equal_to_predict_future():
    df = pd.DataFrame(
        {
            "game_id": [1, 1, 2, 2, 3, 3, 4, 4, 5, 5],
            "player_id": [1, 2, 1, 2, 1, 2, 1, 2, 1, 2],
            "team_id": [1, 2, 1, 2, 1, 2, 1, 2, 1, 2],
            "start_date": [
                pd.to_datetime("2023-01-01"),
                pd.to_datetime("2023-01-01"),
                pd.to_datetime("2023-01-02"),
                pd.to_datetime("2023-01-02"),
                pd.to_datetime("2023-01-03"),
                pd.to_datetime("2023-01-03"),
                pd.to_datetime("2023-01-03"),
                pd.to_datetime("2023-01-03"),
                pd.to_datetime("2023-01-04"),
                pd.to_datetime("2023-01-04"),
            ],
            "deaths": [0.5, 1, 0.7, 2, 0.5, 0.7, 0.2, 2.1, 0.8, 1],
            "kills": [0.2, 0.3, 0.2, 0.3, 0.2, 0.3, 0.2, 0.3, 0.2, 0.3],
            "__target": [1, 0, 0.6, 0.3, 0.8, 0.2, 0.4, 0.1, 1, 0],
        }
    )

    column_weights = [ColumnWeight(name="kills", weight=1)]
    predictor = Predictor(estimator=LinearRegression())

    column_names = ColumnNames(
        match_id="game_id",
        team_id="team_id",
        player_id="player_id",
        start_date="start_date",
    )
    rating_generator = UpdateRatingGenerator(
        known_features_out=[RatingKnownFeatures.RATING_DIFFERENCE_PROJECTED]
    )

    pipeline = Pipeline(
        performances_generator=PerformancesGenerator(
            Performance(weights=column_weights)
        ),
        predictor=predictor,
        rating_generators=rating_generator,
        lag_generators=[
            LagTransformer(
                features=["kills", "deaths"], lag_length=1, granularity=["player_id"]
            )
        ],
        column_names=column_names,
    )

    cross_validator = MatchKFoldCrossValidator(
        match_id_column_name=column_names.match_id,
        n_splits=1,
        min_validation_date="2023-01-04",
        date_column_name=column_names.start_date,
    )
    cross_validated_df = pipeline.train_predict(
        df=df, cross_validate_predict=True, cross_validator=cross_validator
    )

    historical_df = df[df[column_names.start_date] < pd.to_datetime("2023-01-04")]
    future_df = df[df[column_names.start_date] >= pd.to_datetime("2023-01-04")]
    historical_df_predictions = pipeline.train_predict(df=historical_df)
    future_predict = pipeline.future_predict(df=future_df).reset_index(drop=True)
    future_cv_df = cross_validated_df[
        cross_validated_df[column_names.start_date] >= pd.to_datetime("2023-01-04")
    ].reset_index(drop=True)
    past_cv_df = cross_validated_df[
        cross_validated_df[column_names.start_date] < pd.to_datetime("2023-01-04")
    ].reset_index(drop=True)
    pd.testing.assert_frame_equal(
        future_cv_df, future_predict, check_like=True, check_dtype=False
    )
    pd.testing.assert_frame_equal(
        past_cv_df, historical_df_predictions, check_like=True, check_dtype=False
    )


def test_post_pre_and_lag_transformers():
    df = pd.DataFrame(
        {
            "game_id": [1, 1, 2, 2, 3, 3, 4, 4, 5, 5],
            "player_id": [1, 2, 1, 2, 1, 2, 1, 2, 1, 2],
            "team_id": [1, 2, 1, 2, 1, 2, 1, 2, 1, 2],
            "start_date": [
                pd.to_datetime("2023-01-01"),
                pd.to_datetime("2023-01-01"),
                pd.to_datetime("2023-01-02"),
                pd.to_datetime("2023-01-02"),
                pd.to_datetime("2023-01-03"),
                pd.to_datetime("2023-01-03"),
                pd.to_datetime("2023-01-03"),
                pd.to_datetime("2023-01-03"),
                pd.to_datetime("2023-01-04"),
                pd.to_datetime("2023-01-04"),
            ],
            "kills": [0.5, 1, 0.7, 2, 0.5, 0.7, 0.2, 2.1, 0.8, 1],
            "__target": [1, 0, 0.6, 0.3, 0.8, 0.2, 0.4, 0.1, 1, 0],
        }
    )

    column_weights = [ColumnWeight(name="kills", weight=1)]
    predictor = Predictor(estimator=LinearRegression())

    column_names = ColumnNames(
        match_id="game_id",
        team_id="team_id",
        player_id="player_id",
        start_date="start_date",
    )
    rating_generator = UpdateRatingGenerator(
        non_estimator_known_features_out=[RatingKnownFeatures.PLAYER_RATING],
        known_features_out=[RatingKnownFeatures.RATING_DIFFERENCE_PROJECTED],
        historical_features_out=[RatingHistoricalFeatures.TEAM_RATING],
    )
    pre_transformer = PredictorTransformer(
        predictor=Predictor(
            estimator=LinearRegression(),
            estimator_features=rating_generator.features_out,
        ),
    )

    lag_generator = LagTransformer(
        features=["kills"], lag_length=1, granularity=["player_id"]
    )
    post_transformer = RatioTeamPredictorTransformer(
        features=["kills"],
        predictor=Predictor(
            estimator=LinearRegression(), estimator_features=lag_generator.features_out
        ),
    )

    pipeline = Pipeline(
        performances_generator=PerformancesGenerator(
            Performance(weights=column_weights)
        ),
        predictor=predictor,
        rating_generators=rating_generator,
        pre_lag_transformers=[pre_transformer],
        lag_generators=[lag_generator],
        post_lag_transformers=[post_transformer],
        column_names=column_names,
    )
    train_df = df[df[column_names.start_date] < pd.to_datetime("2023-01-04")]
    future_df = df[df[column_names.start_date] >= pd.to_datetime("2023-01-04")]
    trained_df = pipeline.train_predict(train_df, return_features=True)

    for f in rating_generator.known_features_out:
        assert f in trained_df.columns
        assert f not in train_df.columns

    for f in lag_generator.estimator_features_out:
        assert f in trained_df.columns
        assert f not in train_df.columns

    for f in post_transformer.estimator_features_out:
        assert f in trained_df.columns
        assert f not in train_df.columns

    for f in pre_transformer.estimator_features_out:
        assert f in trained_df.columns
        assert f not in train_df.columns

    assert predictor.pred_column in trained_df.columns
    assert predictor.pred_column not in train_df.columns

    predicted_df = pipeline.future_predict(future_df, return_features=True)
    for f in rating_generator.known_features_out:
        assert f in predicted_df.columns
        assert f not in future_df.columns

    for f in lag_generator.estimator_features_out:
        assert f in predicted_df.columns
        assert f not in future_df.columns

    for f in rating_generator.historical_features_out:
        assert f in predicted_df.columns
        assert f not in future_df.columns

    for f in post_transformer.estimator_features_out:
        assert f in predicted_df.columns
        assert f not in future_df.columns

    for f in pre_transformer.estimator_features_out:
        assert f in predicted_df.columns
        assert f not in future_df.columns

    for f in rating_generator.non_estimator_known_features_out:
        assert f in predicted_df.columns
        assert f not in future_df.columns

    assert predictor.pred_column in predicted_df.columns
    assert predictor.pred_column not in future_df.columns



from player_performance_ratings.ratings.performance_generator.performances_transformers import (
    MinMaxTransformer,
    DiminishingValueTransformer,
    SklearnEstimatorImputer,
    PartialStandardScaler,
    SymmetricDistributionTransformer,
    GroupByTransformer,
)
from player_performance_ratings.ratings.performance_generator.performances_generator import (
    ColumnWeight,
    Performance,
    auto_create_pre_performance_transformations,
    PerformancesGenerator,
)



import pandas as pd

from player_performance_ratings import ColumnNames, PredictColumnNames
from player_performance_ratings.pipeline_transformer import PipelineTransformer
from player_performance_ratings.ratings import (
    UpdateRatingGenerator,
    MatchRatingGenerator,
    StartRatingGenerator,
    RatingKnownFeatures,
)
from player_performance_ratings.transformers import LagTransformer
from player_performance_ratings.transformers.lag_generators import (
    RollingMeanTransformerPolars,
)

column_names = ColumnNames(
    team_id="teamname",
    match_id="gameid",
    start_date="date",
    player_id="playername",
    league="league",
    position="position",
)
df = pd.read_parquet("data/subsample_lol_data")
df = (
    df.loc[lambda x: x.position != "team"]
    .assign(team_count=df.groupby("gameid")["teamname"].transform("nunique"))
    .loc[lambda x: x.team_count == 2]
    .assign(
        player_count=df.groupby(["gameid", "teamname"])["playername"].transform(
            "nunique"
        )
    )
    .loc[lambda x: x.player_count == 5]
)
df = df.assign(team_count=df.groupby("gameid")["teamname"].transform("nunique")).loc[
    lambda x: x.team_count == 2
]


# Pretends the last 10 games are future games. The most will be trained on everything before that.
most_recent_10_games = df[column_names.match_id].unique()[-10:]
historical_df = df[~df[column_names.match_id].isin(most_recent_10_games)]
future_df = df[df[column_names.match_id].isin(most_recent_10_games)].drop(
    columns=["result"]
)

rating_generator = UpdateRatingGenerator(
    known_features_out=[RatingKnownFeatures.RATING_DIFFERENCE_PROJECTED],
    performance_column="result",
)

lag_generators = [
    LagTransformer(
        features=["kills", "deaths", "result"], lag_length=3, granularity=["playername"]
    ),
    RollingMeanTransformerPolars(
        features=["kills", "deaths", "result"],
        window=20,
        min_periods=1,
        granularity=["playername"],
    ),
]


transformer = PipelineTransformer(
    column_names=column_names,
    rating_generators=rating_generator,
    lag_generators=lag_generators,
)

historical_df = transformer.fit_transform(historical_df)

future_df = transformer.transform(future_df)
print(future_df.head())
